#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  menu.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = ["BasicMenu","Menu"]

import weakref
import inspect

from .layer import Layer
from .util import ActionDispatcher

class BasicMenu(ActionDispatcher):
    """
    Menu base class without layer support.
    
    Each menu is separated from the other menus and can be switched between at any time.
    
    Actions supported by default:
    
    ``enter`` is triggered everytime the :py:meth:`on_enter()` method has been called.
    
    ``exit`` is triggered everytime the :py:meth:`on_exit()` method has been called.
    
    .. seealso::
       
       See :py:class:`Menu()` for more information.
    
    """
    
    def __init__(self,name,window,peng):
        self.name = name
        self.window = window
        self.peng = peng
        
        self.eventHandlers = {}
        
        self.worlds = []
    def draw(self):
        """
        This method is called if it is time to render the menu.
        
        Override this method in subclasses to customize behavior and actually draw stuff.
        """
        pass
    
    def addWorld(self,world):
        """
        Adds the given world to the internal list.
        
        Worlds that are registered via this method will get all events that are given to this menu passed through.
        
        This mechanic is mainly used to implement actor controllers.
        """
        self.worlds.append(world)
    
    # Event handlers
    def handleEvent(self,event_type,args):
        if event_type in self.eventHandlers:
            for whandler in self.eventHandlers[event_type]:
                # This allows for proper collection of deleted handler methods by using weak references
                handler = whandler()
                if handler is None:
                    del self.eventHandlers[event_type][self.eventHandlers[event_type].index(whandler)]
                handler(*args)
        for world in self.worlds:
            world.handle_event(event_type,args,self.window)
    handleEvent.__noautodoc__ = True
    def registerEventHandler(self,event_type,handler):
        if event_type not in self.eventHandlers:
            self.eventHandlers[event_type]=[]
        # Only a weak reference is kept
        if inspect.ismethod(handler):
            handler = weakref.WeakMethod(handler)
        else:
            handler = weakref.ref(handler)
        self.eventHandlers[event_type].append(handler)
    
    def on_enter(self,old):
        """
        This fake event handler will be called every time this menu is entered via the :py:meth:`PengWindow.changeMenu()` method.
        
        This handler will not be called if this menu is already active.
        """
        pass # Custom fake event handler for entering the menu
    def on_exit(self,new):
        """
        This fake event handler will be called every time this menu is exited via the :py:meth:`PengWindow.changeMenu()` method.
        
        This handler will not be called if this menu is the same as the new menu.
        """
        pass # Custom fake event handler for leaving the menu

class Menu(BasicMenu):
    """
    Subclass of :py:class:`BasicMenu` adding support for the :py:class:`Layer` class.
    
    This subclass overrides the draw and __init__ method, so be sure to call them if you override them.
    """
    
    def __init__(self,name,window,peng):
        super(Menu,self).__init__(name,window,peng)
        self.layers = []
    def addLayer(self,layer,z=-1):
        """
        Adds a new layer to the stack, optionally at the specified z-value.
        
        ``layer`` must be an instance of Layer or subclasses.
        
        ``z`` can be used to override the index of the layer in the stack. Defaults to ``-1`` for appending.
        """
        # Adds a new layer to the stack, optionally at the specified z-value
        # The z-value is the index this layer should be inserted in, or -1 for appending
        if not isinstance(layer,Layer):
            raise TypeError("layer must be an instance of Layer!")
        if z==-1:
            self.layers.append(layer)
        else:
            self.layers.insert(z,layer)
    def draw(self):
        """
        Draws the layers in the appropriate order.
        
        Layers that have their ``enabled`` attribute set to ``False`` are skipped.
        """
        # Draws each layer in the given order
        for layer in self.layers:
            if layer.enabled:
                layer._draw()
    
    def on_enter(self,old):
        """
        Same as :py:meth:`BasicMenu.on_enter()`\ , but also calls :py:meth:`Layer.on_menu_enter()` on every layer.
        """
        for layer in self.layers:
            layer.on_menu_enter(old)
    def on_exit(self,new):
        """
        Same as :py:meth:`BasicMenu.on_exit()`\ , but also calls :py:meth:`Layer.on_menu_exit()` on every layer.
        """
        for layer in self.layers:
            layer.on_menu_exit(new)
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  world.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = ["World","StaticWorld",
           "WorldView", "WorldViewMouseRotatable"]

import weakref
import inspect

from .camera import Camera

try:
    import pyglet
    from pyglet.gl import *
    from pyglet.window import key
    _have_pyglet = True
except ImportError:
    _have_pyglet = False

class World(object):
    """
    World containing terrain, actors, cameras and views.
    
    See the docs about :py:class:`Camera()`\ , :py:class:`WorldView()`\ , :py:class:`Actor()` for more information about each class.
    
    This class does not draw anything, see :py:class:`StaticWorld()` for drawing simple terrain.
    """
    def __init__(self,peng):
        self.peng = peng
        self.cameras = {}
        self.actors = {}
        self.views = {}
        
        self.eventHandlers = {}
        self.recvEvents = True
    
    def addCamera(self,camera):
        """
        Add the camera to the internal registry.
        
        Each camera name must be unique, or else only the most recent version will be used. This behavior should not be relied on because some objects may cache objects.
        
        Additionally, only instances of :py:class:`Camera() <peng3d.camera.Camera>` may be used, everything else raises a :py:exc:`TypeError`\ .
        """
        if not isinstance(camera,Camera):
            raise TypeError("camera is not of type Camera!")
        self.cameras[camera.name]=camera
    def addView(self,view):
        """
        Adds the supplied :py:class:`WorldView()` object to the internal registry.
        
        The same restrictions as for cameras apply, e.g. no duplicate names.
        
        Additionally, only instances of :py:class:`WorldView()` may be used, everything else raises a :py:exc:`TypeError`\ .
        """
        if not isinstance(view,WorldView):
            raise TypeError("view is not of type WorldView!")
        self.views[view.name]=view
    def addActor(self,actor):
        """
        Adds the given actor to the internal registry.
        
        Note that this actors :py:attr:`uuid` attribute must be unique, else it will override any actors previously registered with its UUID.
        """
        self.actors[actor.uuid]=actor
    
    def getView(self,name):
        """
        Returns the view with name ``name``\ .
        
        Raises a :py:exc:`ValueError` if the view does not exist.
        """
        if name not in self.views:
            raise ValueError("Unknown world view")
        return self.views[name]
    
    def render3d(self,view=None):
        """
        Renders the world in 3d-mode.
        
        If you want to render custom terrain, you may override this method. Be careful that you still call the original method or else actors may not be rendered.
        """
        for actor in self.actors.values():
            actor.render(view)
    
    # Event Handlers
    def handle_event(self,event_type,args,window=None):
        if not self.recvEvents:
            if self.peng.cfg["debug.events.dump"]:
                print("World skyps event type %s"%event_type)
            return
        args = list(args)
        #if window is not None:
        #    args.append(window)
        if event_type in self.eventHandlers:
            for whandler in self.eventHandlers[event_type]:
                # This allows for proper collection of deleted handler methods by using weak references
                handler = whandler()
                if handler is None:
                    del self.eventHandlers[event_type][self.eventHandlers[event_type].index(whandler)]
                handler(*args)
    handle_event.__noautodoc__ = True
    def registerEventHandler(self,event_type,handler):
        if self.peng.cfg["debug.events.register"]:
            print("Registered Event: %s Handler: %s"%(event_type,handler))
        if event_type not in self.eventHandlers:
            self.eventHandlers[event_type]=[]
        # Only a weak reference is kept
        if inspect.ismethod(handler):
            handler = weakref.WeakMethod(handler)
        else:
            handler = weakref.ref(handler)
        self.eventHandlers[event_type].append(handler)

class StaticWorld(World):
    """
    Subclass of :py:class:`StaticWorld()`\ , allows for semi-static terrain to be rendered.
    
    This class is not suitable for highly complex or user-modifiable terrain.
    
    ``quads`` is a list of 3d vertices, e.g. a single quad may be ``[-1,-1,-1, 1,-1,-1, 1,-1,1, -1,-1,1]``\ , which represents a rectangle of size 2x2 centered around 0,0.
    It should also be noted that all quads have to be in a single list.
    
    ``colors`` is a list of RGB Colors in a similar format to ``quads`` but with colors instead. Note that there must be a color for every vertex in the vertex list.
    Every color is an integer between 0 and 255 using the internal pyglet scheme ``c3B/static``\ .
    
    You can modify the terrain via the ``terrain`` attribute, note that it is a pyglet vertex list, and not a python list.
    """
    def __init__(self,peng,quads,colors):
        super(StaticWorld,self).__init__(peng)
        assert _have_pyglet
        self.batch3d = pyglet.graphics.Batch()
        self.terrain = self.batch3d.add(len(quads)/3, GL_QUADS, None,
            ("v3f/static",quads),
            ("c3B/static",colors),
            )
    def render3d(self,view=None):
        """
        Renders the world.
        """
        super(StaticWorld,self).render3d(view)
        self.batch3d.draw()

class WorldView(object):
    """
    Object representing a view on the world.
    
    A :py:class:`WorldView()` object references a camera and has a name.
    
    ``cam`` is a valid camera name known to the world object supplied.
    """
    def __init__(self,world,name,cam):
        self.world = world
        self.name = name
        self.activeCamera = cam
        self.cam.on_activate(None)
    
    def setActiveCamera(self,name):
        """
        Sets the active camera.
        
        This method also calls the :py:meth:`Camera.on_activate() <peng3d.camera.Camera.on_activate>` event handler if the camera is not already active.
        """
        if name == self.activeCamera:
            return # Cam is already active
        if name not in self.world.cameras:
            raise ValueError("Unknown camera name")
        old = self.activeCamera
        self.activeCamera = name
        self.cam.on_activate(old)
    
    # Event handlers
    
    def on_menu_enter(self,old):
        """
        Fake event handler called by :py:meth:`Layer.on_menu_enter() <peng3d.layer.Layer.on_menu_enter>` when the containing menu is entered.
        """
        self.world.peng.window.push_handlers(self)
    def on_menu_exit(self,new):
        """
        Fake event handler, same as :py:meth:`on_menu_enter()` but for exiting menus instead.
        """
        self.world.peng.window.pop_handlers()
    
    # Proxy for self.cameras[self.activeCamera]
    @property
    def cam(self):
        """
        Property for getting the currently active camera.
        
        Always equals ``self.cameras[self.activeCamera]``\ .
        """
        return self.world.cameras[self.activeCamera]
    
    # Proxy for self.cam.rot
    @property
    def rot(self):
        """
        Property for accessing the current rotation of the active camera.
        
        This property can also be written to.
        """
        return self.world.cameras[self.activeCamera].rot
    @rot.setter
    def rot(self,value):
        if not (isinstance(value,list) or isinstance(value,tuple)):
            raise TypeError("Invalid type for rot value!")
        self.world.cameras[self.activeCamera].rot = value
    
    # Proxy for self.cam.pos
    @property
    def pos(self):
        """
        Property for accessing the current position of the active camera.
        
        The value of this property will always be a list of length 3.
        
        This property can also be written to.
        """
        return self.world.cameras[self.activeCamera].pos
    @pos.setter
    def pos(self,value):
        if not (isinstance(value,list) or isinstance(value,tuple)):
            raise TypeError("pos must be list or tuple")
        elif len(value)!=3:
            raise ValueError("pos must be of len 3 not len %s"%len(value))
        self.world.cameras[self.activeCamera].pos = list(value)

class WorldViewMouseRotatable(WorldView):
    """
    Subclass of :py:class:`WorldView()` that is rotatable using the user.
    
    Moving the mouse cursor left or right will rotate the attached camera horizontally and moving the mouse cursor up or down will rotate the camera vertically.
    
    By default, each pixel traveled changes the angle in degrees by 0.15, though this can be changed via the :confval:`controls.mouse.sensitivity` config value.
    """
    def on_menu_enter(self,old):
        """
        Fake event handler, same as :py:meth:`WorldView.on_menu_enter()` but forces mouse exclusivity.
        """
        super(WorldViewMouseRotatable,self).on_menu_enter(old)
        self.world.peng.window.toggle_exclusivity(True)
    def on_menu_exit(self,new):
        """
        Fake event handler, same as :py:meth:`WorldView.on_menu_exit()` but force-disables mouse exclusivity.
        """
        super(WorldViewMouseRotatable,self).on_menu_exit(new)
        self.world.peng.window.toggle_exclusivity(False)
    def on_key_press(self,symbol,modifiers):
        """
        Keyboard event handler handling only the escape key.
        
        If an escape key press is detected, mouse exclusivity is toggled via :py:meth:`PengWindow.toggle_exclusivity()`\ .
        """
        if symbol == key.ESCAPE:
            self.world.peng.window.toggle_exclusivity()
            return pyglet.event.EVENT_HANDLED
    def on_mouse_motion(self, x, y, dx, dy):
        """
        Handles mouse motion and rotates the attached camera accordingly.
        
        For more information about how to customize mouse movement, see the class documentation here :py:class:`WorldViewMouseRotatable()`\ .
        """
        if not self.world.peng.window.exclusive:
            return
        m = self.world.peng.cfg["controls.mouse.sensitivity"]
        x, y = self.rot
        x, y = x + dx * m, y + dy * m
        y = max(-90, min(90, y))
        x %= 360
        newrot = (x,y)
        self.rot= newrot
    def on_mouse_drag(self,x,y,dx,dy,buttons,modifiers):
        """
        Handler used to still enable mouse movement while a button is pressed.
        """
        self.on_mouse_motion(x,y,dx,dy)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  keybind.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

import pyglet
from pyglet.window import key

import bidict

MOD_RELEASE = 1 << 15 # Additional fake modifier applied if on_key_release is called
"""
Fake modifier applied when a key is released instead of pressed.

This modifier internally has the value of ``1<<15`` and should thus be safe from any added modifiers in the future.

Note that this modifier is only applied within keybinds, not in regular ``on_key_down`` and ``on_key_up`` handlers.
"""

MODNAME2MODIFIER = bidict.OrderedBidict([
    ("ctrl",key.MOD_ACCEL), # For compat between MacOSX and everything else, equals to MOD_COMMAND on OSX and MOD_CTRL everywhere else
    ("alt",key.MOD_ALT),
    ("shift",key.MOD_SHIFT),
    #("control",key.MOD_ACCEL),
    # Currently disabled due to bidict not allowing duplicate values
    ("option",key.MOD_OPTION), # MacOSX only
    ("capslock",key.MOD_CAPSLOCK),
    ("numlock",key.MOD_NUMLOCK),
    ("scrolllock",key.MOD_SCROLLLOCK),
    #("release",MOD_RELEASE), # Deprecated, use the released flag instead
    ])
"""
Ordered Bidict that maps between user-friendly names and internal constants.

Note that since this is a bidict, you can query the reverse mapping by accessing :py:attr:`MODNAME2MODIFIER.inv`\ .
The non-inverse mapping maps from user-friendly name to internal constant.

This mapping is used by the Keybind system to convert the modifier constants to names.

The Mapping is as follows:

================ =============================== =======
Name             Pyglet constant                 Notes
================ =============================== =======
ctrl             :py:data:`key.MOD_ACCEL`
alt              :py:data:`key.MOD_ALT`          1
shift            :py:data:`key.MOD_SHIFT`
option           :py:data:`key.MOD_OPTION`
capslock         :py:data:`key.MOD_CAPSLOCK`
numlock          :py:data:`key.MOD_NUMLOCK`
scrollock        :py:data:`key.MOD_SCROLLOCK`
================ =============================== =======

1: automatically replaced by MOD_CTRL on Darwin/OSX
"""
if pyglet.compat_platform=="darwin":
    MODNAME2MODIFIER["alt"]=key.MOD_CTRL

OPTIONAL_MODNAMES = ["capslock","numlock","scrollock"]
"""
List of modifiers that are not substantial to a key combo.

If the :confval:`controls.keybinds.strict` option is disabled, every key combo is emitted with and without the modifiers in this list.
Else, only the combo with these modifiers is emitted.

This may cause no more combos to get through if numlock or capslock are activated.
"""

class KeybindHandler(object):
    """
    Handler class that automatically converts incoming key events to key combo events.
    
    A keybinding always is of format ``[MOD1-[MOD2-]]KEY`` with potentially more modifiers.
    
    See :py:data:`MODNAME2MODIFIER` for more information about existing modifiers.
    
    Note that the order in which modifiers are listed also is the order of the above listing.
    
    Keybindings are matched exactly, and optionally a second time without the modifiers listed in :py:data:`OPTIONAL_MODNAMES` if :confval:`controls.keybinds.strict` is set to False.
    """
    def __init__(self,peng):
        self.peng = peng
        self.peng.registerEventHandler("on_key_press",self.on_key_press)
        self.peng.registerEventHandler("on_key_release",self.on_key_release)
        self.keybinds = {}
        self.keybinds_nm = {}
        self.kbname = bidict.bidict()
    def add(self,keybind,kbname,handler,mod=True):
        """
        Adds a keybind to the internal registry.
        
        Keybind names should be of the format ``namespace:category.subcategory.name``\ e.g. ``peng3d:actor.player.controls.forward`` for the forward key combo for the player actor.
        
        :param str keybind: Keybind string, as described above
        :param str kbname: Name of the keybind, may be used to later change the keybinding without re-registering
        :param function handler: Function or any other callable called with the positional arguments ``(symbol,modifiers,release)`` if the keybind is pressed or released
        :param int mod: If the keybind should respect modifiers
        """
        keybind = keybind.lower()
        if mod:
            if keybind not in self.keybinds:
                self.keybinds[keybind]=[]
            self.keybinds[keybind].append(kbname)
        else:
            if keybind not in self.keybinds_nm:
                self.keybinds_nm[keybind]=[]
            self.keybinds_nm[keybind].append(kbname)
        self.kbname[kbname]=handler
        self.peng.sendEvent("peng3d:keybind.add",{"peng":self.peng,"keybind":keybind,"kbname":kbname,"handler":handler,"mod":mod})
    def changeKeybind(self,kbname,combo):
        """
        Changes a keybind of a specific keybindname.
        
        :param str kbname: Same as kbname of :py:meth:`add()`
        :param str combo: New key combination
        """
        for key,value in self.keybinds.items():
            if kbname in value:
                del value[value.index(kbname)]
                break
        if combo not in self.keybinds:
            self.keybinds[combo]=[]
        self.keybinds[combo].append(kbname)
        self.peng.sendEvent("peng3d.keybind.change",{"peng":self.peng,"kbname":kbname,"combo":combo})
    def mod_is_held(self,modname,modifiers):
        """
        Helper method to simplify checking if a modifier is held.
        
        :param str modname: Name of the modifier, see :py:data:`MODNAME2MODIFIER`
        :param int modifiers: Bitmask to check in, same as the modifiers argument of the on_key_press etc. handlers
        """
        modifier = MODNAME2MODIFIER[modname.lower()]
        return modifiers&modifier
    def on_key_press(self,symbol,modifiers,release=False):
        # Format: "MOD1-MOD2-MOD3-KEY" in all lowercase
        modnames = []
        for modname in MODNAME2MODIFIER:
            if self.mod_is_held(modname,modifiers):
                modnames.append(modname)
        keyname = key.symbol_string(symbol)
        modnames.append(keyname)
        combo = "-".join(modnames).lower()
        self.handle_combo(combo,symbol,modifiers,release,True)
        self.handle_combo(keyname.lower(),symbol,modifiers,release,False)
        if not self.peng.cfg["controls.keybinds.strict"]:
            newmodnames = []
            for modname in modnames:
                if modname not in OPTIONAL_MODNAMES:
                    newmodnames.append(modname)
            if newmodnames==modnames:
                return
            combo = "-".join(newmodnames).lower()
            self.handle_combo(combo,symbol,modifiers,release,True)
    def handle_combo(self,combo,symbol,modifiers,release=False,mod=True):
        """
        Handles a key combination and dispatches associated events.
        
        First, all keybind handlers registered via :py:meth:`add` will be handled,
        then the pyglet event :peng3d:pgevent:`on_key_combo` with params ``(combo,symbol,modifiers,release,mod)`` is sent to the :py:class:`Peng()` instance.
        
        Also sends the events :peng3d:event:`peng3d:keybind.combo`\, :peng3d:event:`peng3d:keybind.combo.press` and :peng3d:event`peng3d:keybind.combo.release`\ .
        
        :params str combo: Key combination pressed
        :params int symbol: Key pressed, passed from the same argument within pyglet
        :params int modifiers: Modifiers held while the key was pressed
        :params bool release: If the combo was released
        :params bool mod: If the combo was sent without mods
        """
        if self.peng.cfg["controls.keybinds.debug"]:
            print("combo: nm=%s %s"%(mod,combo))
        if mod:
            for kbname in self.keybinds.get(combo,[]):
                self.kbname[kbname](symbol,modifiers,release)
        else:
            for kbname in self.keybinds_nm.get(combo,[]):
                self.kbname[kbname](symbol,modifiers,release)
        self.peng.sendPygletEvent("on_key_combo",(combo,symbol,modifiers,release,mod))
        self.peng.sendEvent("peng3d:keybind.combo",{"peng":self.peng,"combo":combo,"symbol":symbol,"modifiers":modifiers,"release":release,"mod":mod})
        if release:
            self.peng.sendEvent("peng3d:keybind.combo.release",{"peng":self.peng,"combo":combo,"symbol":symbol,"modifiers":modifiers,"release":release,"mod":mod})
        else:
            self.peng.sendEvent("peng3d:keybind.combo.press",{"peng":self.peng,"combo":combo,"symbol":symbol,"modifiers":modifiers,"release":release,"mod":mod})
    def on_key_release(self,symbol,modifiers):
        #self.on_key_press(symbol,modifiers|MOD_RELEASE)
        self.on_key_press(symbol,modifiers,release=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  player.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = ["BasicPlayer","FirstPersonPlayer","FourDirectionalMoveController","EgoMouseRotationalController","BasicFlightController"]

import math

try:
    import pyglet
except ImportError:
    pass

from . import Actor,RotatableActor,Controller

class FourDirectionalMoveController(Controller):
    """
    Controller allowing the user to control the actor with the keyboard.
    
    You can configure the used keybinds with the :confval:`controls.controls.forward` etc.
    The keybinds can also be changed with their ``keybindname``\ , e.g. ``peng3d:actor.<actor uuid>.player.controls.forward`` for forward.
    
    The movement speed may also be changed via the :py:attr:`movespeed` instance attribute, which defaults to :confval:`controls.controls.movespeed`\ .
    
    You may also access the currently held keys via :py:attr:`move`\ , which is a list with 2 items, forwards/backwards and left/right.
    """
    def __init__(self,*args,**kwargs):
        super(FourDirectionalMoveController,self).__init__(*args,**kwargs)
        self.move = [0,0]
        self.movespeed = self.peng.cfg["controls.controls.movespeed"]
    def registerEventHandlers(self):
        """
        Registers needed keybinds and schedules the :py:meth:`update` Method.
        
        You can control what keybinds are used via the :confval:`controls.controls.forward` etc. Configuration Values.
        """
        # Forward
        self.peng.keybinds.add(self.peng.cfg["controls.controls.forward"],"peng3d:actor.%s.player.controls.forward"%self.actor.uuid,self.on_fwd_down,False)
        # Backward
        self.peng.keybinds.add(self.peng.cfg["controls.controls.backward"],"peng3d:actor.%s.player.controls.backward"%self.actor.uuid,self.on_bwd_down,False)
        # Strafe Left
        self.peng.keybinds.add(self.peng.cfg["controls.controls.strafeleft"],"peng3d:actor.%s.player.controls.strafeleft"%self.actor.uuid,self.on_left_down,False)
        # Strafe Right
        self.peng.keybinds.add(self.peng.cfg["controls.controls.straferight"],"peng3d:actor.%s.player.controls.straferight"%self.actor.uuid,self.on_right_down,False)
        pyglet.clock.schedule_interval(self.update,1.0/60)
    def update(self,dt):
        """
        Should be called regularly to move the actor.
        
        This method does nothing if the :py:attr:`enabled` property is set to false.
        
        Note that this method is called automatically and should not be manually called.
        """
        if not self.enabled:
            return
        speed = self.movespeed
        d = dt * speed # distance covered this tick.
        dx, dy, dz = self.get_motion_vector()
        # New position in space, before accounting for gravity.
        dx, dy, dz = dx * d, dy * d, dz * d
        #dy+=self.vert*VERT_SPEED
        x,y,z = self.actor._pos
        newpos = dx+x, dy+y, dz+z
        self.actor.pos = newpos
    def get_motion_vector(self):
        """
        Returns the movement vector according to held buttons and the rotation.
        
        :return: 3-Tuple of ``(dx,dy,dz)``
        :rtype: tuple
        """
        if any(self.move):
            x, y = self.actor._rot
            strafe = math.degrees(math.atan2(*self.move))
            y_angle = math.radians(y)
            x_angle = math.radians(x + strafe)
            dy = 0.0
            dx = math.cos(x_angle)
            dz = math.sin(x_angle)
        else:
            dy = 0.0
            dx = 0.0
            dz = 0.0
        return (dx, dy, dz)
    
    # Event Handlers
    def on_fwd_down(self,symbol,modifiers,release):
        self.move[0]-=1 if not release else -1
    def on_bwd_down(self,symbol,modifiers,release):
        self.move[0]+=1 if not release else -1
    
    def on_left_down(self,symbol,modifiers,release):
        self.move[1]-=1 if not release else -1
    def on_right_down(self,symbol,modifiers,release):
        self.move[1]+=1 if not release else -1

class EgoMouseRotationalController(Controller):
    """
    Controller allowing the user to rotate the actor with the mouse.
    """
    def __init__(self,*args,**kwargs):
        super(EgoMouseRotationalController,self).__init__(*args,**kwargs)
    def registerEventHandlers(self):
        """
        Registers the motion and drag handlers.
        
        Note that because of the way pyglet treats mouse dragging, there is also an handler registered to the on_mouse_drag event.
        """
        self.world.registerEventHandler("on_mouse_motion",self.on_mouse_motion)
        self.world.registerEventHandler("on_mouse_drag",self.on_mouse_drag)
    def on_mouse_motion(self, x, y, dx, dy):
        if self.enabled:
            m = self.peng.cfg["controls.mouse.sensitivity"]
            x, y = self.actor._rot
            x, y = x + dx * m, y + dy * m
            y = max(-90, min(90, y))
            x %= 360
            newrot = (x,y)
            self.actor.rot = newrot
    def on_mouse_drag(self,x,y,dx,dy,buttons,modifiers):
        self.on_mouse_motion(x,y,dx,dy)

class BasicFlightController(Controller):
    """
    Controller allowing the user to move up and down with the jump and crouch controls.
    
    The used keybinds may be configured via :confval:`controls.controls.crouch` and :confval:`controls.controls.jump`\ .
    
    The vertical speed used when flying may be configured via :confval:`controls.controls.verticalspeed` or the :py:attr:`speed` attribute.
    """
    def __init__(self,*args,**kwargs):
        super(BasicFlightController,self).__init__(*args,**kwargs)
        self.speed = self.peng.cfg["controls.controls.verticalspeed"]
        self.move = 0
    def registerEventHandlers(self):
        """
        Registers the up and down handlers.
        
        Also registers a scheduled function every 60th of a second, causing pyglet to redraw your window with 60fps.
        """
        # Crouch/fly down
        self.peng.keybinds.add(self.peng.cfg["controls.controls.crouch"],"peng3d:actor.%s.player.controls.crouch"%self.actor.uuid,self.on_crouch_down,False)
        # Jump/fly up
        self.peng.keybinds.add(self.peng.cfg["controls.controls.jump"],"peng3d:actor.%s.player.controls.jump"%self.actor.uuid,self.on_jump_down,False)
        pyglet.clock.schedule_interval(self.update,1.0/60)
    def update(self,dt):
        """
        Should be called regularly to move the actor.
        
        This method does nothing if the :py:attr:`enabled` property is set to False.
        
        This method is called automatically and should not be called manually.
        """
        if not self.enabled:
            return
        dy = self.speed * dt * self.move
        x,y,z = self.actor._pos
        newpos = x,dy+y,z
        self.actor.pos = newpos
    
    def on_crouch_down(self,symbol,modifiers,release):
        self.move -= 1 if not release else -1
    def on_jump_down(self,symbol,modifiers,release):
        self.move += 1 if not release else -1

class BasicPlayer(RotatableActor):
    """
    Basic Player class, subclass of :py:class:`RotatableActor()`\ .
    
    This class adds no features currently, it can be used to identify player actors via :py:func:`isinstance()`\ .
    """
    pass

class FirstPersonPlayer(BasicPlayer):
    """
    Old class allowing to create standard first-person players easily.
    
    :deprecated: See :py:class:`EgoMouseRotationalController()` and :py:class:`FourDirectionalMoveController()` instead
    """
    def __init__(self,peng,world,uuid=None,pos=[0,0,0],rot=[0,0]):
        super(FirstPersonPlayer,self).__init__(peng,world,uuid,pos,rot)
        self.move = [0,0]
        self.movespeed = self.peng.cfg["controls.controls.movespeed"]
        self.active = True
        
        # Event handler registration
        # Forward
        self.peng.keybinds.add(self.peng.cfg["controls.controls.forward"],"peng3d:actor.player.controls.forward",self.on_fwd_down)
        self.peng.keybinds.add("release-"+self.peng.cfg["controls.controls.forward"],"peng3d:actor.player.controls.forward.release",self.on_fwd_up)
        # Backward
        self.peng.keybinds.add(self.peng.cfg["controls.controls.backward"],"peng3d:actor.player.controls.backward",self.on_bwd_down)
        self.peng.keybinds.add("release-"+self.peng.cfg["controls.controls.backward"],"peng3d:actor.player.controls.backward.release",self.on_bwd_up)
        # Strafe Left
        self.peng.keybinds.add(self.peng.cfg["controls.controls.strafeleft"],"peng3d:actor.player.controls.strafeleft",self.on_left_down)
        self.peng.keybinds.add("release-"+self.peng.cfg["controls.controls.strafeleft"],"peng3d:actor.player.controls.strafeleft.release",self.on_left_up)
        # Strafe Right
        self.peng.keybinds.add(self.peng.cfg["controls.controls.straferight"],"peng3d:actor.player.controls.straferight",self.on_right_down)
        self.peng.keybinds.add("release-"+self.peng.cfg["controls.controls.straferight"],"peng3d:actor.player.controls.straferight.release",self.on_right_up)
        # Mouse
        self.world.registerEventHandler("on_mouse_motion",self.on_mouse_motion)
        self.world.registerEventHandler("on_mouse_drag",self.on_mouse_drag)
        pyglet.clock.schedule_interval(self.update,1.0/60)
    def update(self,dt):
        """
        Internal method used for moving the player.
        
        :param float dt: Time delta since the last call to this method
        """
        speed = self.movespeed
        d = dt * speed # distance covered this tick.
        dx, dy, dz = self.get_motion_vector()
        # New position in space, before accounting for gravity.
        dx, dy, dz = dx * d, dy * d, dz * d
        #dy+=self.vert*VERT_SPEED
        x,y,z = self._pos
        newpos = dx+x, dy+y, dz+z
        self.pos = newpos
    def get_motion_vector(self):
        """
        Returns the movement vector according to held buttons and the rotation.
        
        :return: 3-Tuple of ``(dx,dy,dz)``
        :rtype: tuple
        """
        if any(self.move):
            x, y = self._rot
            strafe = math.degrees(math.atan2(*self.move))
            y_angle = math.radians(y)
            x_angle = math.radians(x + strafe)
            dy = 0.0
            dx = math.cos(x_angle)
            dz = math.sin(x_angle)
        else:
            dy = 0.0
            dx = 0.0
            dz = 0.0
        return (dx, dy, dz)
    
    # Event Handlers
    def on_fwd_down(self,symbol,modifiers):
        self.move[0]-=1
    def on_fwd_up(self,symbol,modifiers):
        self.move[0]+=1
    
    def on_bwd_down(self,symbol,modifiers):
        self.move[0]+=1
    def on_bwd_up(self,symbol,modifiers):
        self.move[0]-=1
    
    def on_left_down(self,symbol,modifiers):
        self.move[1]-=1
    def on_left_up(self,symbol,modifiers):
        self.move[1]+=1
    
    def on_right_down(self,symbol,modifiers):
        self.move[1]+=1
    def on_right_up(self,symbol,modifiers):
        self.move[1]-=1
    
    def on_mouse_motion(self, x, y, dx, dy):
        if self.active:
            m = 0.15
            x, y = self._rot
            x, y = x + dx * m, y + dy * m
            y = max(-90, min(90, y))
            x %= 360
            newrot = (x,y)
            self.rot = newrot
    def on_mouse_drag(self,x,y,dx,dy,buttons,modifiers):
        self.on_mouse_motion(x,y,dx,dy)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  text.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = [
    "Label",
    "TextInput","TextInputBackground"
    ]

import time

import pyglet
from pyglet.gl import *
from pyglet.window import key

try:
    import pyperclip
    HAVE_PYPERCLIP = True
except ImportError:
    HAVE_PYPERCLIP = False

from .widgets import Background,Widget,mouse_aabb
from .button import ButtonBackground


class Label(Widget):
    """
    Simple widget that can display any single-line non-formatted string.
    
    This widget does not use any background by default.
    
    The default font color is chosen to work on the default background color and may need to be changed if the background color is changed.
    """
    def __init__(self,name,submenu,window,peng,
                 pos=None,size=None,
                 bg=None,
                 label="Label",
                 font_size=16,font="Arial",
                 font_color=[62,67,73,255],
                 multiline=False,
                 label_cls=pyglet.text.Label,
                ):
        super(Label,self).__init__(name,submenu,window,peng,pos,size,bg)
        
        self.font_name = font
        self.font_size = font_size
        self.font_color = font_color
        
        if label_cls == pyglet.text.HTMLLabel:
            self._label = label_cls(str(label),
                    # Font is changed later
                    x=0,y=0,
                    batch=self.submenu.batch2d,
                    anchor_x="center", anchor_y="center",
                    group=pyglet.graphics.OrderedGroup(1),
                    width=self.size[0],height=self.size[1],
                    multiline=multiline,
                    )
            self._label.font_name = font
            self._label.font_size = font_size
            self._label.font_color = font_color
        else:
            self._label = label_cls(str(label),
                    font_name=font,
                    font_size=font_size,
                    color=font_color,
                    x=0,y=0,
                    batch=self.submenu.batch2d,
                    anchor_x="center", anchor_y="center",
                    group=pyglet.graphics.OrderedGroup(1),
                    width=self.size[0],height=self.size[1],
                    multiline=multiline,
                    )
        if getattr(label,"_dynamic",False):
            def f():
                self.label = str(label)
            self.peng.i18n.addAction("setlang",f)
        
        self.redraw()
    
    def on_redraw(self,dt=None):
        super(Label,self).on_redraw()
        self.redraw_label()
    def redraw_label(self):
        """
        Re-draws the text by calculating its position.
        
        Currently, the text will always be centered on the position of the label.
        """
        # Convenience variables
        sx,sy = self.size
        x,y = self.pos
        
        # Label position
        self._label.font_name = self.font_name
        self._label.font_size = self.font_size
        self._label.font_color = self.font_color
        
        self._label.x = int(x+sx/2.)
        self._label.y = int(y+sy/2.)
        self._label.width = self.size[0]
        self._label.height = self.size[1]
        self._label._update() # Needed to prevent the label from drifting to the top-left after resizing by odd amounts
    
    @property
    def label(self):
        """
        Property for accessing the text of the label.
        """
        return self._label.text
    @label.setter
    def label(self,label):
        self._label.text = str(label)


class TextInputBackground(ButtonBackground):
    """
    Background for the :py:class:`TextInput` Widget.
    
    This background uses the button drawing routines and adds a cursor.
    """
    def __init__(self,*args,**kwargs):
        self.stime = 0
        super(TextInputBackground,self).__init__(*args,**kwargs)
    def init_bg(self):
        super(TextInputBackground,self).init_bg()
        self.vlist_cursor = self.submenu.batch2d.add(2,GL_LINES,pyglet.graphics.OrderedGroup(10),
            "v2f",
            "c3B",
            )
        self.reg_vlist(self.vlist_cursor)
    def redraw_bg(self):
        super(TextInputBackground,self).redraw_bg()
        
        sx,sy,x,y,bx,by = self.getPosSize()
        
        # TODO: make this less hacky
        otext = self.widget._text.text
        self.widget._text.text = self.widget._text.text[:self.widget.cursor_pos]
        tw = self.widget._text.content_width+2 if len(self.widget._text.text)!=0 else 0
        self.widget._text.text = otext
        
        v = x+tw+bx,y+by, x+tw+bx,y+sy-by
        self.vlist_cursor.vertices = v
        
        bg,o,i,s,h = self.getColors()
        # TODO: add compat with arbitrary backgrounds, will not work on material
        c = s*2 if (self.stime-time.time())%1>.5 or not self.widget.focussed else [0,0,0]*2
        self.vlist_cursor.colors = c
    
    @property
    def pressed(self):
        return self.change_on_press and self.widget.focussed


class TextInput(Widget):
    """
    Basic Textual Input widget.
    
    By default, this widget uses :py:class:`TextInputBackground` as its Background class.
    
    The optional default text will only be displayed if the text is empty.
    
    The ``allow_overflow`` flag determines if the text entered can be longer than the size of the :py:class:`TextInput`\ .
    
    The ``allow_copypaste`` flag controls whether or not the user can copy and paste the contents of the text box.
    By default, copying and pasting is allowed. This flag can also be set to ``"force"`` to force a crash
    with an appropriate error message if the :py:mod:`pyperclip` module is not available.
    Currently, only copying, pasting and cutting the whole text box is supported, as there is no mechanism for text selection yet.
    
    The key combinations used by this widget can be configured in the config via the :confval:`controls.keybinds.common.*` config values.
    """
    def __init__(self,name,submenu,window,peng,
                 pos=None,size=None,
                 bg=None,
                 text="",default="",
                 border=[4,4],borderstyle="flat",
                 font_size=16,font="Arial",
                 font_color=[62,67,73,255],
                 font_color_default=[62,67,73,200],
                 allow_overflow=False,
                 allow_copypaste=True,
                 min_size=None,
                 ):
        
        if allow_copypaste == "force" and not HAVE_PYPERCLIP:
            raise ValueError("%s with name %s requires Clipboard support, but Pyperclip is not installed"%(self.__class__.__name__,name))
        
        if bg is None:
            bg = TextInputBackground(self,border,borderstyle)
        super(TextInput,self).__init__(name,submenu,window,peng,pos,size,bg,min_size)
        
        self.cursor_pos = 0
        self.focussed = False
        self.allow_overflow = allow_overflow
        self.allow_copypaste = True if allow_copypaste is True or allow_copypaste=="force" else False
        self.force_copypaste = True if allow_copypaste=="force" else False
        
        self._text = pyglet.text.Label(str(text),
                font_name=font,
                font_size=font_size,
                color=font_color,
                x=0,y=0,
                batch=None,#self.submenu.batch2d,
                anchor_x="left", anchor_y="center",
                width=self.size[0],height=self.size[1]
                )
        self._default = pyglet.text.Label(str(default),
                font_name=font,
                font_size=font_size,
                color=font_color_default,
                x=0,y=0,
                batch=None,
                anchor_x="left", anchor_y="center",
                width=self.size[0],height=self.size[1]
                )
        self.cursor_pos = 0
        self.focussed = False
        
        if getattr(text,"_dynamic",False):
            def f():
                self.text = str(text)
            self.peng.i18n.addAction("setlang",f)
        if getattr(default,"_dynamic",False):
            def f():
                self.default = str(default)
            self.peng.i18n.addAction("setlang",f)
        
        self.peng.i18n.addAction("setlang",self.redraw) # for dynamic size
        
        self.peng.registerEventHandler("on_text",self.on_text)
        self.peng.registerEventHandler("on_text_motion",self.on_text_motion)
        if self.allow_copypaste:
            self.peng.keybinds.add(self.window.cfg["controls.keybinds.common.copy"],"peng3d:gui.widget.textinput.%s.copy"%self.name,self.on_copy)
            self.peng.keybinds.add(self.window.cfg["controls.keybinds.common.paste"],"peng3d:gui.widget.textinput.%s.paste"%self.name,self.on_paste)
            self.peng.keybinds.add(self.window.cfg["controls.keybinds.common.cut"],"peng3d:gui.widget.textinput.%s.cut"%self.name,self.on_cut)
        
        self.redraw()
        pyglet.clock.schedule_interval(lambda dt: self.redraw(),1./2.)
    
    def on_redraw(self):
        super(TextInput,self).on_redraw()
        self.redraw_label()
    def redraw_label(self):
        """
        Re-draws the label by calculating its position.
        
        Currently, the label will always be centered on the position of the label.
        """
        # Convenience variables
        sx,sy = self.size
        x,y = self.pos
        
        # Label position
        x = x+self.bg.border[0]
        y = y+sy/2.-self._text.font_size/4.
        w = self.size[0]
        h = self.size[1]
        
        self._text.x,self._text.y = x,y
        self._text.width,self._text.height=w,h
        
        self._default.x,self._default.y = x,y
        self._default.width,self._default.height=w,h
        
        self._text._update() # Needed to prevent the label from drifting to the top-left after resizing by odd amounts
        self._default._update()
    
    def draw(self):
        super(TextInput,self).draw()
        if self._text.text=="":
            self._default.draw()
        else:
            self._text.draw()
    draw.__noautodoc__ = True
    
    def on_text(self,text):
        if not (self.focussed and self.clickable):
            return
        
        otext = self.text
        t = self.text
        t = t[:self.cursor_pos]+text+t[self.cursor_pos:]
        self.text = t
        self.cursor_pos+=len(self.text)-len(otext)
        self.cursor_pos = min(self.cursor_pos,len(self.text))
        self.redraw()
    
    def on_text_motion(self,motion):
        if not (self.focussed and self.clickable):
            return
        
        if motion == key.MOTION_BACKSPACE:
            l = list(self.text)
            try:
                # Deletes the character in front of the cursor
                del l[self.cursor_pos-1]
            except Exception:
                return
            self.text = "".join(l)
            self.cursor_pos-=1
            self.cursor_pos = max(self.cursor_pos,0)
            self.redraw()
        elif motion == key.MOTION_DELETE:
            l = list(self.text)
            try:
                # Deletes the character after the cursor
                del l[self.cursor_pos]
            except Exception:
                return
            self.text = "".join(l)
            self.redraw()
        elif motion == key.MOTION_LEFT:
            # Moves the cursor to the left one character
            self.cursor_pos-=1
            self.cursor_pos = max(self.cursor_pos,0)
            self.redraw()
        elif motion == key.MOTION_RIGHT:
            # Moves the cursor to the right one character
            self.cursor_pos+=1
            self.cursor_pos = min(self.cursor_pos,len(self.text))
            self.redraw()
        elif motion == key.MOTION_BEGINNING_OF_LINE or motion == key.MOTION_BEGINNING_OF_FILE:
            # Moves the cursor to the beginning
            self.cursor_pos=0
            self.redraw()
        elif motion == key.MOTION_END_OF_LINE or motion == key.MOTION_END_OF_FILE:
            # Moves the cursor to the end
            self.cursor_pos = len(self.text)
            self.redraw()
    
    def on_mouse_press(self,x,y,button,modifiers):
        if not self.clickable:
            return
        elif mouse_aabb([x,y],self.size,self.pos):
            if button == pyglet.window.mouse.LEFT:
                self.doAction("press")
                self.pressed = True
                self.focussed = True
                self.bg.stime = time.time()
            elif button == pyglet.window.mouse.RIGHT:
                self.doAction("context")
            self.redraw()
        else:
            self.focussed = False
            self.redraw()
    
    def on_copy(self,symbol,modifiers,release):
        if release:
            return
        
        try:
            pyperclip.copy(self.text)
        except Exception:
            if self.force_copypaste:
                raise
    def on_paste(self,symbol,modifiers,release):
        if release:
            return
        
        try:
            t = pyperclip.paste()
            
            # Tries to copy as much as possible, until it does not fit
            if not self.allow_overflow:
                # TODO: implement more efficient algorithm
                self.text = ""
                for c in t:
                    self.text+=c
            else:
                self.text=t
        except Exception:
            if self.force_copypaste:
                raise
    def on_cut(self,symbol,modifiers,release):
        if release:
            return
        
        try:
            pyperclip.copy(self.text)
            self.text = ""
        except Exception:
            if self.force_copypaste:
                raise
    
    @property
    def text(self):
        """
        Property for accessing the text.
        """
        return self._text.text
    @text.setter
    def text(self,text):
        text = str(text)
        otext = self._text.text
        self._text.text = text
        self._text._update()
        # Reverts the change if
        # 1. the text length has NOT increased
        # 2. and allow_overflow is False
        # 3. and the text is wider than the widget size minus the border size
        # 4. and the text is not empty
        if len(otext)<len(text) and not self.allow_overflow and self.size[0]-self.bg.border[0]*2<=self._text.content_width if len(self.text)!=0 else 0:
            self._text.text=otext
            self._text._update()
        self.doAction("textchange")
        self.redraw() # necessary for size/pos that depends on label size

    def getContentSize(self):
        l = [max(self._text.content_width,self._default.content_width),max(self._text.content_height,self._default.content_height)]
        b = self.bg.border # TODO: make this work with borderless backgrounds
        return [l[0]+b[0]*2,l[1]+b[1]*2]
    
    @property
    def default(self):
        """
        Property for accessing the default text.
        """
        return self._default.text
    @default.setter
    def default(self,default):
        self._default.text = str(default)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  layer.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = ["Layer", "Layer2D", "Layer3D","LayerGroup","LayerWorld"]

try:
    import pyglet
except ImportError:
    pass

from . import camera

class Layer(object):
    """
    Base layer class.
    
    A Layer can be used to separate background from foreground or the 3d world from a 2d HUD.
    
    This class by itself does nothing, you will need to subclass it and override the :py:meth:`draw()` method.
    """
    
    def __init__(self,menu,window,peng):
        self.window = window
        self.menu = menu
        self.peng = peng
        self.enabled = True
    
    # Subclass overrides
    def draw(self):
        """
        Called when this layer needs to be drawn.
        
        Override this method in subclasses to redefine behavior.
        """
        pass
    def predraw(self):
        """
        Called before the :py:meth:`draw()` method is called.
        
        This method is used in the :py:class:`Layer2D()` and :py:class:`Layer3D()` subclasses for setting OpenGL state.
        
        Override this method in subclasses to redefine behavior.
        """
        pass
    def postdraw(self):
        """
        Called after the :py:meth:`draw()` method is called.
        
        This method can be used to reset OpenGL state to avoid conflicts with other code.
        
        Override this method in subclasses to redefine behavior.
        """
        pass
    # End subclass overrides
    
    # Event handlers
    
    def on_menu_enter(self,old):
        """
        Custom fake event handler called by :py:meth:`Menu.on_enter()` for every layer.
        
        Useful for adding and removing event handlers per layer.
        """
        pass
    def on_menu_exit(self,new):
        """
        Custom fake event handler called by :py:meth:`Menu.on_exit()` for every layer.
        
        Useful for adding and removing event handlers per layer.
        """
        pass
    
    def _draw(self):
        if not self.enabled:
            return
        self.predraw()
        try:
            self.draw()
        except Exception:
            raise
        finally:
            self.postdraw()

class Layer2D(Layer):
    """
    2D Variant of :py:class:`Layer()` and a subclass of the former.
    
    This class makes use of the :py:meth:`predraw()` method to configure OpenGL to draw 2-Dimensionally.
    This class uses :py:meth:`PengWindow.set2d()` to set the 2D mode.
    
    When overriding the :py:meth:`predraw()` method, make sure to call the superclass.
    """
    def predraw(self):
        """
        Uses :py:meth:`PengWindow.set2d()` to enable a 2D OpenGL state.
        """
        self.window.set2d()

class Layer3D(Layer):
    """
    3D Variant of :py:class:`Layer()` and a subclass of the former.
    
    This class works the same as :py:class:`Layer2D()`\ , only for 3D drawing instead.
    This class uses :py:meth:`PengWindow.set3d()` to set the 3D mode.
    
    Also, the correct :py:func:`glTranslatef()` and :py:func:`glRotatef()` are applied to simplify drawing objects.
    When writing the :py:meth:`draw()` method of this class, you will only need to use world coordinates, not camera coordinates.
    This allows for easy building of Games using First-Person-Perspectives.
    """
    def __init__(self,menu,window,peng,cam):
        super(Layer3D,self).__init__(menu,window,peng)
        if not isinstance(cam,camera.Camera):
            raise TypeError("cam must be of type Camera!")
        self.cam = cam
    def predraw(self):
        """
        Uses :py:meth:`PengWindow.set3d()` to enable a 3D OpenGL state.
        """
        self.window.set3d(self.cam)

class LayerGroup(Layer):
    """
    Layer variant wrapping the supplied pyglet group.
    
    ``group`` may only be an instance of :py:class:`pyglet.graphics.Group`\ , else a :py:exc:`TypeError` will be raised.
    
    Also note that both the :py:meth:`predraw() <Layer.predraw()>` and :py:meth:`postdraw() <Layer.postdraw()>` methods are overwritten by this class.
    
    .. seealso::
       
       For more information about pyglet groups, see `the pyglet docs <http://pyglet.readthedocs.io/en/latest/programming_guide/graphics.html#setting-the-opengl-state>`_\ .
    
    """
    def __init__(self,menu,window,peng,group):
        super(LayerGroup,self).__init__(menu,window,peng)
        if not isinstance(group,pyglet.graphics.Group):
            raise TypeError("group must be an instance of pyglet.graphics.Group")
        self.group = group
    def predraw(self):
        """
        Sets the group state.
        """
        self.group.set_state()
    def postdraw(self):
        """
        Re-sets the previous state.
        """
        self.group.unset_state()

class LayerWorld(Layer3D):
    """
    Subclass of :py:class:`Layer3D()` implementing a 3D Layer showing a specific :py:class:`WorldView`\ .
    
    All arguments passed to the constructor should be self-explanatory.
    
    Note that you may not set any of the attributes directly, or crashes and bugs may appear indirectly within a certain during future re-drawing of the screen.
    """
    def __init__(self,menu,window,peng,world,viewname):
        super(LayerWorld,self).__init__(menu,window,peng,world.getView(viewname).cam)
        self.world = world
        self.viewname = viewname
        self.view = self.world.getView(self.viewname)
    def setView(self,name):
        """
        Sets the view used to the specified ``name``\ .
        
        The name must be known to the world or else a :py:exc:`ValueError` is raised.
        """
        if name not in self.world.views:
            raise ValueError("Invalid viewname for world!")
        self.viewname = viewname
        self.view = self.world.getView(self.viewname)
    def predraw(self):
        """
        Sets up the attributes used by :py:class:`Layer3D()` and calls :py:meth:`Layer3D.predraw()`\ .
        """
        self.cam = self.view.cam
        super(LayerWorld,self).predraw()
    def draw(self):
        """
        Draws the view using the :py:meth:`World.render3d()` method.
        """
        self.world.render3d(self.view)
    def on_menu_enter(self,old):
        """
        Passes the event through to the WorldView to allow for custom behavior.
        """
        return self.view.on_menu_enter(old)
    def on_menu_exit(self,new):
        """
        Same as :py:meth:`on_menu_enter()`\ .
        """
        return self.view.on_menu_exit(new)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  gui.py
#  
#  Copyright 2017 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = [
    "mouse_aabb",
    "points2htmlfontsize",
    "ResourceGroup",
    "BlitableTexture",
    ]

import pyglet
from pyglet.gl import *

def mouse_aabb(mpos,size,pos):
    """
    AABB Collision checker that can be used for most axis-aligned collisions.
    
    Intended for use in widgets to check if the mouse is within the bounds of a particular widget.
    """
    return pos[0]<=mpos[0]<=pos[0]+size[0] and pos[1]<=mpos[1]<=pos[1]+size[1]

def points2htmlfontsize(points):
    """
    Approximate font size converter, converts from Points to HTML ``<font>`` tag font sizes.
    
    Note that this method is very inaccurate, since there are only seven possible output values that represent at least 25 input values.
    When in doubt, this function always rounds down, e.g. every input value less than eight is converted to HTML size 1.
    """
    # Approximation, always rounds down
    if points<=8:
        return 1
    elif points<=10:
        return 2
    elif points<=12:
        return 3
    elif points<=14:
        return 4
    elif points<=18:
        return 5
    elif points<=24:
        return 6
    else:
        # Inaccurate, represents everything above 24, but should equal 48
        return 7

class ResourceGroup(pyglet.graphics.Group):
    """
    Pyglet Group that represents a Resource as returned by the :py:class:`ResourceManager() <peng3d.resource.ResourceManager>`\ .
    
    This Group should automatically merge different groups with different resources that are on the same texture atlas.
    """
    def __init__(self,data,parent=None):
        super(ResourceGroup, self).__init__(parent)
        self.data = data

    def set_state(self):
        glEnable(self.data[0])
        glBindTexture(self.data[0], self.data[1])
    set_state.__noautodoc__ = True

    def unset_state(self):
        glDisable(self.data[0])
    unset_state.__noautodoc__ = True

    def __hash__(self):
        return hash((self.data[0], self.data[1], self.parent))

    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
            self.data[0] == other.data[0] and
            self.data[1] == other.data[1] and
            self.parent == other.parent)

    def __repr__(self):
        return '%s(id=%d)' % (self.__class__.__name__, self.data[1])

class BlitableTexture(pyglet.image.Texture):
    """
    Texture subclass allowing :py:class:`pyglet.image.atlas.TextureBin()` to copy a texture to itself.
    
    This class simply implements the :py:meth:`blit_to_texture()` interface.
    """
    # Basic Texture subclass allowing pyglet.image.atlas.TextureBin to copy a texture to itself
    def blit_to_texture(self,target,level,x,y,z=0):
        glCopyTexSubImage2D(target, level, x,y, 0,0, self.width,self.height)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  pyglet_patch.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  
"""
These patches are used to fix specific problems with pyglet.

Note that these are not likely to stay, since I will have to verify that they still work with every Pyglet Release.
"""

__all__ = ["patch_float2int"]

import sys
import ctypes

import pyglet
import pyglet.graphics
from pyglet.graphics.vertexbuffer import MappableVertexBufferObject
from pyglet.graphics import vertexbuffer

from pyglet.gl import *

### patch_float2int

def _get_region(self, buffer, start, count):
        '''Map a buffer region using this attribute as an accessor.

        The returned region can be modified as if the buffer was a contiguous
        array of this attribute (though it may actually be interleaved or
        otherwise non-contiguous).

        The returned region consists of a contiguous array of component
        data elements.  For example, if this attribute uses 3 floats per
        vertex, and the `count` parameter is 4, the number of floats mapped
        will be ``3 * 4 = 12``.

        :Parameters:
            `buffer` : `AbstractMappable`
                The buffer to map.
            `start` : int
                Offset of the first vertex to map.
            `count` : int
                Number of vertices to map

        :rtype: `AbstractBufferRegion`
        '''
        byte_start = self.stride * start
        byte_size = self.stride * count
        array_count = self.count * count
        if self.stride == self.size or not array_count:
            # non-interleaved
            ptr_type = ctypes.POINTER(self.c_type * array_count)
            return buffer.get_region(byte_start, byte_size, ptr_type)
        else:
            # interleaved
            byte_start += self.offset
            byte_size -= self.offset
            elem_stride = self.stride // ctypes.sizeof(self.c_type)
            elem_offset = self.offset // ctypes.sizeof(self.c_type)
            ptr_type = ctypes.POINTER(
                self.c_type * int((count * elem_stride - elem_offset)))
            region = buffer.get_region(byte_start, byte_size, ptr_type)
            return vertexbuffer.IndirectArrayRegion(
                region, array_count, self.count, elem_stride)

def _bind(self):
        # Commit pending data
        super(MappableVertexBufferObject, self).bind()
        size = self._dirty_max - self._dirty_min
        if size > 0:
            if size == self.size:
                glBufferData(self.target, self.size, self.data, self.usage)
            else:
                glBufferSubData(self.target, self._dirty_min, int(size),
                    self.data_ptr + self._dirty_min)
            self._dirty_min = sys.maxsize
            self._dirty_max = 0

def _iar__setitem__(self, index, value):
        count = self.count
        if not isinstance(index, slice):
            elem = index // count
            j = index % count
            self.region.array[elem * self.stride + j] = value
            return

        start = index.start or 0
        stop = index.stop
        step = index.step or 1
        if start < 0:
            start = self.size + start
        if stop is None:
            stop = self.size
        elif stop < 0:
            stop = self.size + stop

        assert step == 1 or step % count == 0, \
            'Step must be multiple of component count'

        data_start = (start // count) * self.stride + start % count
        data_stop = (stop // count) * self.stride + stop % count

        # ctypes does not support stepped slicing, so do the work in a list
        # and copy it back.
        data = self.region.array[:]
        if step == 1:
            data_step = self.stride
            value_step = count
            for i in range(count):
                data[data_start + i:int(data_stop) + i:data_step] = \
                    value[i::value_step]
        else:
            data_step = (step // count) * self.stride
            data[data_start:data_stop:data_step] = value
        self.region.array[:] = data

def _draw(self, mode, vertex_list=None):
        '''Draw vertices in the domain.

        If `vertex_list` is not specified, all vertices in the domain are
        drawn.  This is the most efficient way to render primitives.

        If `vertex_list` specifies a `VertexList`, only primitives in that
        list will be drawn.

        :Parameters:
            `mode` : int
                OpenGL drawing mode, e.g. ``GL_POINTS``, ``GL_LINES``, etc.
            `vertex_list` : `VertexList`
                Vertex list to draw, or ``None`` for all lists in this domain.

        '''
        glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT)
        for buffer, attributes in self.buffer_attributes:
            buffer.bind()
            for attribute in attributes:
                attribute.enable()
                attribute.set_pointer(attribute.buffer.ptr)
        if vertexbuffer._workaround_vbo_finish:
            glFinish()

        if vertex_list is not None:
            glDrawArrays(mode, vertex_list.start, vertex_list.count)
        else:
            starts, sizes = self.allocator.get_allocated_regions()
            primcount = len(starts)
            if primcount == 0:
                pass
            elif primcount == 1:
                # Common case
                glDrawArrays(mode, starts[0], int(sizes[0]))
            elif gl_info.have_version(1, 4):
                starts = (GLint * primcount)(*starts)
                sizes = (GLsizei * primcount)(*sizes)
                glMultiDrawArrays(mode, starts, sizes, primcount)
            else:
                for start, size in zip(starts, sizes):
                    glDrawArrays(mode, start, size)

        for buffer, _ in self.buffer_attributes:
            buffer.unbind()
        glPopClientAttrib()

def patch_float2int():
    """
    Patches the :py:mod:`pyglet.graphics.vertexattribute`\ , :py:mod:`pyglet.graphics.vertexbuffer` and :py:mod:`pyglet.graphics.vertexdomain` modules.
    
    This patch is only needed with Python 3.x and will be applied automatically when initializing :py:class:`Peng()`\ .
    
    The patches consist of simply converting some list indices, slices and other numbers to integers from floats with .0. These patches have not been tested thoroughly, but work with at least ``test.py`` and ``test_gui.py``\ .
    
    Can be enabled and disabled via :confval:`pyglet.patch.patch_float2int`\ .
    """
    pyglet.graphics.vertexattribute.AbstractAttribute.get_region = _get_region
    pyglet.graphics.vertexbuffer.MappableVertexBufferObject.bind = _bind
    pyglet.graphics.vertexbuffer.IndirectArrayRegion.__setitem__ = _iar__setitem__
    pyglet.graphics.vertexdomain.VertexDomain.draw = _draw

### End patch_float2int



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  layered.py
#  
#  Copyright 2017 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = [
    "LayeredWidget",
    "BasicWidgetLayer","WidgetLayer",
    "GroupWidgetLayer",
    "ImageWidgetLayer","DynImageWidgetLayer",
    "ImageButtonWidgetLayer",
    "LabelWidgetLayer",
    "FormattedLabelWidgetLayer","HTMLLabelWidgetLayer",
    "BaseBorderWidgetLayer","ButtonBorderWidgetLayer",
    ]

import warnings

import pyglet
from pyglet.gl import *

from .widgets import Background, Widget
from .. import util

class LayeredWidget(Widget):
    """
    Layered Widget allowing for easy creation of custom widgets.
    
    A Layered Widget consists of (nearly) any amount of layers in a specific order.
    
    All Layers should be subclasses of :py:class:`BasicWidgetLayer` or :py:class:`WidgetLayer`\ .
    
    ``layers`` must be a list of 2-tuples of ``(layer,z_index)``\ .
    """
    def __init__(self,name,submenu,window,peng,
                pos=None,size=None,
                bg=None,layers=[],
                ):
        super(LayeredWidget,self).__init__(name,submenu,window,peng,pos,size,bg)
        
        self.layers = []
        self._layers = {}
        for l,z in layers:
            # Makes sure that all layers are sorted
            # Does not use sort() to keep original order
            self.addLayer(l,z)
    def addLayer(self,layer,z_index=None):
        """
        Adds the given layer at the given Z Index.
        
        If ``z_index`` is not given, the Z Index specified by the layer will be used.
        """
        if z_index is None:
            z_index = layer.z_index
        i = 0
        for l,z in self.layers:
            if z>z_index:
                break
            i+=1
        self._layers[layer.name]=layer
        self.layers.insert(i,[layer,z_index])
    def getLayer(self,name):
        """
        Returns the layer corresponding to the given name.
        
        :raises KeyError: If there is no Layer with the given name.
        """
        return self._layers[name]
    def on_redraw(self):
        super(LayeredWidget,self).on_redraw()
        for layer,_ in self.layers:
            layer.on_redraw()
    def redraw_layer(self,name):
        """
        Redraws the given layer.
        
        :raises ValueError: If there is no Layer with the given name.
        """
        if name not in self._layers:
            raise ValueError("Layer %s not part of widget, cannot redraw")
        self._layers[name].on_redraw()
    def draw(self):
        """
        Draws all layers of this LayeredWidget.
        
        This should normally be unneccessary, since it is recommended that layers use Vertex Lists instead of OpenGL Immediate Mode.
        """
        super(LayeredWidget,self).draw()
        for layer,_ in self.layers:
            layer._draw()
    def delete(self):
        """
        Deletes all layers within this LayeredWidget before deleting itself.
        
        Recommended to call if you are removing the widget, but not yet exiting the interpreter.
        """
        for layer,_ in self.layers:
            layer.delete()
        self.layers = []
        self._layers = {}
        super(LayeredWidget,self).delete()

class BasicWidgetLayer(object):
    """
    Base class for all Layers to be used with :py:class:`LayeredWidget()`\ .
    
    Not to be confused with :py:class:`peng3d.layer.Layer()`\ , these classes are not compatible.
    
    It is recommended to use :py:class:`WidgetLayer()` instead, since functionality is limited in this basic class.
    
    Note that the ``z_index`` will default to a reasonable value for most subclasses and thus is not required to be given explicitly.
    The ``z_index`` for this Layer defaults to ``0``\ .
    """
    z_index = 0
    def __init__(self,name,widget,
                z_index=None,
                ):
        self.name = name
        self.widget = widget
        
        if z_index is not None:
            self.z_index = z_index
        self.group = pyglet.graphics.OrderedGroup(self.z_index)
        
        self._vlists = []
    
    def on_redraw(self):
        """
        Called by the parent widget if this Layer should be redrawn.
        
        Note that it is recommended to call the Baseclass Variant of this Method first when overwriting it.
        See :py:meth:`WidgetLayer.on_redraw` for more information.
        """
        pass
    def _draw(self):
        self.predraw()
        self.draw()
        self.postdraw()
    
    def predraw(self):
        """
        Called before calling the :py:meth:`draw()` Method.
        
        Useful for setting up OpenGL state.
        """
        pass
    def draw(self):
        """
        Called to draw the layer.
        
        Note that using this function is discouraged, use Pyglet Vertex Lists instead.
        
        If you want to call this method manually, call :py:meth:`_draw()` instead.
        This will make sure that :py:meth:`predraw()` and :py:meth:`postdraw()` are called.
        """
        pass
    def postdraw(self):
        """
        Called after calling the :py:meth:`draw()` Method.
        
        Useful for unsetting OpenGL state.
        """
        pass
    
    def regVList(self,vlist):
        """
        Registers a vertex list for proper deletion once this Layer gets destroyed.
        
        This prevents visual artifacts from forming during deletion of a layer.
        """
        self._vlists.append(vlist)
    def delete(self):
        """
        Deletes this Layer.
        
        Currently only deletes VertexLists registered with :py:meth:`regVList()`\ .
        """
        for vlist in self._vlists:
            vlist.delete()
        self._vlists = []

class WidgetLayer(BasicWidgetLayer):
    """
    Subclass of :py:class:`WidgetLayer()` adding commonly used utility features.
    
    This subclass adds a border and offset system.
    
    The ``border`` is a 2-tuple of ``(x_border,y_border)``\ . The border is applied to all sides, resulting in the size being decreased by two pixel per pixel border width.
    
    ``offset`` is relative to the bottom left corner of the screen.
    """
    def __init__(self,name,widget,
                z_index=None,
                border=[0,0],offset=[0,0],
                ):
        super(WidgetLayer,self).__init__(name,widget,z_index)
        
        self._initialized = False
        self._border = border
        self._offset = offset
    def on_redraw(self):
        """
        Called when the Layer should be redrawn.
        
        If a subclass uses the :py:meth:`initialize()` Method, it is very important to also call the Super Class Method to prevent crashes.
        """
        super(WidgetLayer,self).on_redraw()
        if not self._initialized:
            self.initialize()
            self._initialized = True
    def initialize(self):
        """
        Called just before :py:meth:`on_redraw()` is called the first time.
        """
        pass
    
    @property
    def border(self):
        """
        Property to be used for setting and getting the border of the layer.
        
        Note that setting this property causes an immediate redraw.
        """
        if callable(self._border):
            return util.WatchingList(self._border(*(self.widget.pos+self.widget.size)),self._wlredraw_border)
        else:
            return util.WatchingList(self._border,self._wlredraw_border)
    @border.setter
    def border(self,value):
        self._border = value
        self.on_redraw()
    
    @property
    def offset(self):
        """
        Property to be used for setting and getting the offset of the layer.
        
        Note that setting this property causes an immediate redraw.
        """
        if callable(self._offset):
            return util.WatchingList(self._offset(*(self.widget.pos+self.widget.size)),self._wlredraw_offset)
        else:
            return util.WatchingList(self._offset,self._wlredraw_offset)
    @offset.setter
    def offset(self,value):
        self._offset = value
        self.on_redraw()
    
    def _wlredraw_border(self,wl):
        self.border = wl[:]
    def _wlredraw_offset(self,wl):
        self.offset = wl[:]
    
    def getPos(self):
        """
        Returns the absolute position and size of the layer.
        
        This method is intended for use in vertex position calculation, as the border and offset have already been applied.
        
        The returned value is a 4-tuple of ``(sx,sy,ex,ey)``\ .
        The two values starting with an s are the "start" position, or the lower-left corner.
        The second pair of values signify the "end" position, or upper-right corner.
        """
        # Returns sx,sy,ex,ey
        # sx,sy are bottom-left/lowest
        # ex,ey are top-right/highest
        sx,sy = self.widget.pos[0]+self.border[0]+self.offset[0],                       self.widget.pos[1]+self.border[1]+self.offset[1]
        ex,ey = self.widget.pos[0]+self.widget.size[0]-self.border[0]+self.offset[0],   self.widget.pos[1]+self.widget.size[1]-self.border[1]+self.offset[1]
        return sx,sy,ex,ey
    def getSize(self):
        """
        Returns the size of the layer, with the border size already subtracted.
        """
        return self.widget.size[0]-self.border[0]*2,self.widget.size[1]-self.border[1]*2

class GroupWidgetLayer(WidgetLayer):
    """
    Subclass of :py:class:`WidgetLayer()` allowing for using a pyglet group to manage OpenGL state.
    
    If no pyglet group is given, :py:class:`pyglet.graphics.NullGroup()` will be used.
    """
    def __init__(self,name,widget,
                group=None,z_index=None,
                border=[0,0],offset=[0,0],
                ):
        super(GroupWidgetLayer,self).__init__(name,widget,z_index,border,offset)
        
        self.group = group if group is not None else pyglet.graphics.NullGroup()
    def predraw(self):
        self.group.set_state()
    predraw.__noautodoc__ = True
    def postdraw(self):
        self.group.unset_state()
    postdraw.__noautodoc__ = True

class ImageWidgetLayer(WidgetLayer):
    """
    Subclass of :py:class:`WidgetLayer()` implementing a simple static image view.
    
    This layer can display any resource representable by the :py:class:`ResourceManager()`\ .
    
    ``img`` is a 2-tuple of ``(resource_name,category)``\ .
    
    The ``z_index`` for this Layer defaults to ``1``\ .
    """
    z_index = 1
    def __init__(self,name,widget,
                z_index=None,
                border=[0,0],offset=[0,0],
                img=[None,None],
                ):
        super(ImageWidgetLayer,self).__init__(name,widget,z_index,border,offset)
        
        self.img = widget.peng.resourceMgr.getTex(*img)
        self.img_group = util.ResourceGroup(self.img,self.group)
    
    def initialize(self):
        self.img_vlist = self.widget.submenu.batch2d.add(4,GL_QUADS,self.img_group,
                        "v2f",
                        ("t3f",self.img[2]),
                        )
        self.regVList(self.img_vlist)
    initialize.__noautodoc__ = True
    def on_redraw(self):
        super(ImageWidgetLayer,self).on_redraw()
        sx,sy, ex,ey = self.getPos()
        
        # A simple rectangle
        self.img_vlist.vertices =  sx,sy, ex,sy, ex,ey, sx,ey

class _DynImageGroup(pyglet.graphics.Group):
    def __init__(self,layer,parent=None):
        super(_DynImageGroup,self).__init__(parent)
        self.layer = layer
    def set_state(self):
        tex_info = self.layer.imgs[self.layer.cur_img]
        glEnable(tex_info[0])
        glBindTexture(tex_info[0],tex_info[1])
    def unset_state(self):
        tex_info = self.layer.imgs[self.layer.cur_img]
        glDisable(tex_info[0])

class DynImageWidgetLayer(WidgetLayer):
    """
    Subclass of :py:class:`WidgetLayer` allowing for dynamic images.
    
    ``imgs`` is a dictionary of names to 2-tuples of ``(resource_name,category)``\ .
    
    If no default image name is given, a semi-random one will be selected.
    
    The ``z_index`` for this Layer defaults to ``1``\ .
    """
    z_index = 1
    def __init__(self,name,widget,
                z_index=None,
                border=[0,0],offset=[0,0],
                imgs={},
                default=None,
                ):
        super(DynImageWidgetLayer,self).__init__(name,widget,z_index,border,offset)
        
        self.imgs = {}
        self.img_group = _DynImageGroup(self,self.group)
        for name,rsrc in imgs.items():
            self.addImage(name,rsrc)
        
        self.cur_img = None
        self.default_img = default
    
    def addImage(self,name,rsrc):
        """
        Adds an image to the internal registry.
        
        ``rsrc`` should be a 2-tuple of ``(resource_name,category)``\ .
        """
        self.imgs[name]=self.widget.peng.resourceMgr.getTex(*rsrc)
    def switchImage(self,name):
        """
        Switches the active image to the given name.
        
        :raises ValueError: If there is no such image
        """
        if name not in self.imgs:
            raise ValueError("No image of name '%s'"%name)
        elif self.cur_img==name:
            return
        
        self.cur_img = name
        self.on_redraw()
    
    def initialize(self):
        self.img_vlist = self.widget.submenu.batch2d.add(4,GL_QUADS,self.img_group,
                        "v2f",
                        "t3f",
                        )
        self.regVList(self.img_vlist)
        
        self.cur_img = self.cur_img if self.cur_img is not None else (self.default_img if self.default_img is not None else list(self.imgs.keys())[0])
    initialize.__noautodoc__ = True
    def on_redraw(self):
        super(DynImageWidgetLayer,self).on_redraw()
        sx,sy, ex,ey = self.getPos()
        
        # A simple rectangle
        self.img_vlist.vertices =  sx,sy, ex,sy, ex,ey, sx,ey
        
        self.img_vlist.tex_coords = self.imgs[self.cur_img][2]

class ImageButtonWidgetLayer(DynImageWidgetLayer):
    """
    Subclass of :py:class:`DynImageWidgetLayer()` that acts like an :py:class:`ImageButton()`\ .
    
    The ``img_*`` arguments are of the same format as in :py:class:`DynImageWidgetLayer()`\ .
    
    This class internally uses the :py:meth:`BasicWidget.getState()` method for getting the state of the widget.
    """
    def __init__(self,name,widget,
                z_index=None,
                border=[0,0],offset=[0,0],
                img_idle=None,img_pressed=None,img_hover=None,img_disabled=None,
                ):
        super(ImageButtonWidgetLayer,self).__init__(name,widget,z_index,border,
                    imgs={"idle":img_idle,
                          "pressed":img_pressed,
                          "hover":img_hover,
                          "disabled":img_disabled,
                          },
                    default="idle",
                    )
        self.widget.addAction("statechanged",lambda:self.switchImage(self.widget.getState()))

class LabelWidgetLayer(WidgetLayer):
    """
    Subclass of :py:class:`WidgetLayer()` displaying arbitrary plain text.
    
    Note that this method internally uses a pyglet Label that is centered on the Layer.
    
    The ``z_index`` for this Layer defaults to ``2``\ .
    """
    z_index = 2
    def __init__(self,name,widget,
                z_index=None,
                border=[0,0],offset=[0,0],
                label="",
                font_size=16,font="Arial",
                font_color=[62,67,73,255],
                multiline=False,
                ):
        super(LabelWidgetLayer,self).__init__(name,widget,z_index,border,offset)
        
        self._label = pyglet.text.Label(str(label),
                font_name=font,
                font_size=font_size,
                color=font_color,
                x=0,y=0,
                batch=self.widget.submenu.batch2d,
                group=self.group,
                anchor_x="center", anchor_y="center",
                width=self.getSize()[0],#height=self.getSize()[1],
                multiline=multiline,
                )
        if getattr(label,"_dynamic",False):
            def f():
                self.label = str(label)
            self.peng.i18n.addAction("setlang",f)
        
        self.on_redraw()
    
    def on_redraw(self,dt=None):
        super(LabelWidgetLayer,self).on_redraw()
        self.redraw_label()
    def redraw_label(self):
        """
        Re-draws the text by calculating its position.
        
        Currently, the text will always be centered on the position of the layer.
        """
        # Convenience variables
        x,y,_,_ = self.getPos()
        sx,sy = self.getSize()
        
        self._label.x = x+sx/2.
        self._label.y = y+sy/2.
        self._label.width = sx
        # Height is not set, would look weird otherwise
        #self._label.height = sx
        self._label._update() # Needed to prevent the label from drifting to the top-left after resizing by odd amounts
    
    @property
    def label(self):
        """
        Property for accessing the text of the label.
        """
        return self._label.text
    @label.setter
    def label(self,label):
        self._label.text = label

class FormattedLabelWidgetLayer(WidgetLayer):
    """
    Subclass of :py:class:`WidgetLayer()` serving as a base class for other formatted label layers.
    
    The Label Type can be set via the class attribute ``cls``\ , it should be set to any class that is compatible with :py:class:`pyglet.text.Label`\ .
    
    It is recommended to use one of the subclasses of this class instead of this class directly.
    
    The ``z_index`` for this Layer defaults to ``2``\ .
    """
    z_index = 2
    cls = pyglet.text.HTMLLabel
    def __init__(self,name,widget,
                z_index=None,
                border=[0,0],offset=[0,0],
                label="",
                font_size=None,font="Arial",
                font_color=None,
                multiline=False,
                ):
        super(FormattedLabelWidgetLayer,self).__init__(name,widget,z_index,border,offset)
        
        self.font_size = font_size
        self.font_name = font
        self.font_color = font_color
        
        self._label = self.cls(str(label),
                x=0,y=0,
                batch=self.widget.submenu.batch2d,
                group=self.group,
                anchor_x="center", anchor_y="center",
                width=self.getSize()[0],#height=self.getSize()[1],
                multiline=multiline,
                )
        if getattr(label,"_dynamic",False):
            def f():
                self.label = str(label)
            self.peng.i18n.addAction("setlang",f)
        
        self.on_redraw()
    
    def on_redraw(self,dt=None):
        super(FormattedLabelWidgetLayer,self).on_redraw()
        self.redraw_label()
    def redraw_label(self):
        """
        Re-draws the text by calculating its position.
        
        Currently, the text will always be centered on the position of the layer.
        """
        # Convenience variables
        x,y,_,_ = self.getPos()
        sx,sy = self.getSize()
        
        if self.font_name is not None:
            self._label.font_name = self.font_name
        if self.font_size is not None:
            self._label.font_size = self.font_size
        if self.font_color is not None:
            self._label.color = self.font_color
        
        self._label.x = x+sx/2.
        self._label.y = y+sy/2.
        self._label.width = sx
        # Height is not set, would look weird otherwise
        #self._label.height = sx
        self._label._update() # Needed to prevent the label from drifting to the top-left after resizing by odd amounts
    
    @property
    def label(self):
        """
        Property for accessing the text of the label.
        
        Note that depending on the type of format, this property may not exactly represent the original text as it is converted internally.
        """
        return self._label.text
    @label.setter
    def label(self,label):
        self._label.text = label

class HTMLLabelWidgetLayer(FormattedLabelWidgetLayer):
    """
    Subclass of :py:class:`FormattedLabelWidgetLayer` implementing a basic HTML Label.
    
    Note that not all tags are supported, see the docs for :py:class:`pyglet.text.HTMLLabel` for details.
    """
    cls = pyglet.text.HTMLLabel

# TODO: support other formats (Markdown, Pyglet-style Attributed Text, Maybe others?)

class BaseBorderWidgetLayer(WidgetLayer):
    """
    Subclass of :py:class:`WidgetLayer` that displays a basic border around the layer.
    
    Note that not all styles will look good with this class, see :py:class:`ButtonBorderWidgetLayer()` for more information.
    
    Note that the ``border`` and ``offset`` arguments have been renamed to ``base_border`` and ``base_offset`` to prevent naming conflicts.
    
    Subclasses may set the :py:attr:`n_vertices` value to change the number of
    vertices or :py:attr:`change_on_press` to change the default value for the
    argument of the same name.
    By default, 36 vertices are used and ``changed_on_press`` is set to ``True``\ .
    
    The ``z_index`` for this Layer defaults to ``0.5``\ .
    """
    z_index = 0.5 # between default and image layer
    
    n_vertices = 36
    change_on_press = True
    
    color_bg = None
    color_o = None
    color_i = None
    color_s = None
    color_h = None
    
    def __init__(self,name,widget,
                z_index=None,
                base_border=[0,0],base_offset=[0,0],
                border=[4,4, 4,4, 4,4, 4,4],
                style="flat",
                batch=None, change_on_press=None,
                ):
        super(BaseBorderWidgetLayer,self).__init__(name,widget,z_index,base_border,base_offset)
        
        self.bborder = border
        self.style = style
        
        if style=="material" and self.__class__==BaseBorderWidgetLayer:
            warnings.warn("Material style may have visual artifacts if used with BaseBorderWidgetLayer, use ButtonBorderWidgetLayer instead",stacklevel=2)
        elif style=="gradient" and self.__class__==BaseBorderWidgetLayer:
            warnings.warn("Gradient style may have visual artifacts if used with BaseBorderWidgetLayer, use ButtonBorderWidgetLayer instead",stacklevel=2)
        elif style=="oldshadow" and self.__class__==BaseBorderWidgetLayer:
            warnings.warn("Oldshadow style may have visual artifacts if used with BaseBorderWidgetLayer, use ButtonBorderWidgetLayer instead",stacklevel=2)
        
        self.batch = batch
        
        self.change_on_press = change_on_press if change_on_press is not None else self.change_on_press
        
        self.styles = {}
        self.addStyle("flat",self.s_flat)
        self.addStyle("gradient",self.s_gradient)
        self.addStyle("oldshadow",self.s_oldshadow)
        self.addStyle("material",self.s_material)
    
    def addStyle(self,name,func):
        """
        Adds a style to the layer.
        
        Note that styles must be registered seperately for each layer.
        
        ``name`` is the (string) name of the style.
        
        ``func`` will be called with its arguments as ``(bg,o,i,s,h)``\ , see :py:meth:`getColors()` for more information.
        """
        self.styles[name]=func
    def getColors(self):
        """
        Overrideable function that generates the colors to be used by various styles.
        
        Should return a 5-tuple of ``(bg,o,i,s,h)``\ .
        
        ``bg`` is the base color of the background.
        
        ``o`` is the outer color, it is usually the same as the background color.
        
        ``i`` is the inner color, it is usually lighter than the background color.
        
        ``s`` is the shadow color, it is usually quite a bit darker than the background.
        
        ``h`` is the highlight color, it is usually quite a bit lighter than the background.
        
        The returned values may also be statically overridden by setting the :py:attr:`color_<var>` attribute to anything but ``None``\ .
        """
        
        bg = self.widget.submenu.bg[:3] if isinstance(self.widget.submenu.bg,list) or isinstance(self.widget.submenu.bg,tuple) else [242,241,240]
        bg = bg if self.color_bg is None else self.color_bg
        
        o,i = bg, [min(bg[0]+8,255),min(bg[1]+8,255),min(bg[2]+8,255)]
        s,h = [max(bg[0]-40,0),max(bg[1]-40,0),max(bg[2]-40,0)], [min(bg[0]+12,255),min(bg[1]+12,255),min(bg[2]+12,255)]
        
        o = o if self.color_o is None else self.color_o
        i = i if self.color_i is None else self.color_i
        s = s if self.color_s is None else self.color_s
        h = h if self.color_h is None else self.color_h
        
        # Outer,Inner,Shadow,Highlight
        return bg,o,i,s,h
    
    def initialize(self):
        self.batch = self.batch if self.batch is not None else self.widget.submenu.batch2d
        
        self.vlist = self.batch.add(self.n_vertices,GL_QUADS,self.group,
            "v2f",
            "c3B",
            )
        self.regVList(self.vlist)
    initialize.__noautodoc__ = True
    def on_redraw(self):
        super(BaseBorderWidgetLayer,self).on_redraw()
        self.vlist.vertices = self.genVertices()
        
        if self.style not in self.styles:
            raise ValueError("Invalid Style")
        c = self.styles[self.style](*self.getColors())
        
        if len(c)==self.n_vertices*3: # one color per vertex
            # No need to change anything
            pass
        elif len(c)==20*3: # old button-style coloring
            c = self.stretchColors(c)
        else:
            raise ValueError("Style produced %s colors, but %s required"%(len(c)/3,self.n_vertices))
        
        self.vlist.colors = c
    def genVertices(self):
        """
        Called to generate the vertices used by this layer.
        
        The length of the output of this method should be three times the :py:attr:`n_vertices` attribute.
        
        See the source code of this method for more information about the order of the vertices.
        """
        sx,sy,ex,ey = self.getPos()
        b = self.bborder
        
        # Vertex Naming
        # Y
        # |1  2  3  4
        # |5  6  7  8
        # |9  10 11 12
        # |13 14 15 16
        # +------> X
        # Border order
        # 4 2-tuples
        # Each marks x,y offset from the respective corner
        # tuples are in order topleft,topright,bottomleft,bottomright
        # indices:
        # 0,1:topleft; 2,3:topright; 4,5:bottomleft; 6,7:bottomright
        # For a simple border that is even, just repeat the first tuple three more times
        
        v1 = sx,            ey
        v2 = sx+b[0],       ey
        v3 = ex-b[2],       ey
        v4 = ex,            ey
        
        v5 = sx,            ey-b[1]
        v6 = sx+b[0],       ey-b[1]
        v7 = ex-b[2],       ey-b[3]
        v8 = ex,            ey-b[3]
        
        v9 = sx,            sy+b[5]
        v10= sx+b[4],       sy+b[5]
        v11= ex-b[6],       sy+b[7]
        v12= ex,            sy+b[7]
        
        v13= sx,            sy
        v14= sx+b[4],       sy
        v15= ex-b[6],       sy
        v16= ex,            sy
        
        # Layer is separated into 9 sections, naming:
        # 1 2 3
        # 4 5 6
        # 7 8 9
        # Within each section, vertices are given counter-clockwise, starting with the bottom-left
        # 4 3
        # 1 2
        # This is important when assigning colors
        
        q1 = v5 +v6 +v2 +v1
        q2 = v6 +v7 +v3 +v2
        q3 = v7 +v8 +v4 +v3
        q4 = v9 +v10+v6 +v5
        q5 = v10+v11+v7 +v6
        q6 = v11+v12+v8 +v7
        q7 = v13+v14+v10+v9
        q8 = v14+v15+v11+v10
        q9 = v15+v16+v12+v11
        
        return q1+q2+q3+q4+q5+q6+q7+q8+q9
    def stretchColors(self,c):
        """
        Method that is called to stretch the colors.
        
        Note that this should be implemented by subclasses if plausible and reasonable.
        """
        # Subclasses should override this method to implement color stretching
        # Not possible here since the corners are implemented differently
        raise NotImplementedError("Color stretching is not supported on %s"%self.__class__.__name__)
    
    @property
    def pressed(self):
        """
        Read-only helper property to be used by styles for determining if the layer should be rendered as pressed or not.
        
        Note that this property may not represent the actual pressed state, it will always be False if ``change_on_press`` is disabled.
        """
        return self.change_on_press and self.widget.pressed
    @property
    def is_hovering(self):
        """
        Read-only helper property to be used by styles for determining if the layer should be rendered as hovered or not.
        
        Note that this property may not represent the actual hovering state, it will always be False if ``change_on_press`` is disabled.
        """
        return self.change_on_press and self.widget.is_hovering
    
    def s_flat(self,bg,o,i,s,h):
        # Flat style makes no difference between normal,hover and pressed
        return i*36
    s_flat.__noautodoc__ = True
    def s_gradient(self,bg,o,i,s,h):
        if self.pressed:
            i = s
        elif self.is_hovering:
            i = [min(i[0]+6,255),min(i[1]+6,255),min(i[2]+6,255)]
        
        return (
            o+i+o+o+
            i+i+o+o+
            i+o+o+o+
            o+i+i+o+
            i+i+i+i+
            i+o+o+i+
            o+o+i+o+
            o+o+i+i+
            o+o+o+i
            )
    s_gradient.__noautodoc__ = True
    def s_oldshadow(self,bg,o,i,s,h):
        if self.pressed:
            i = s
            s,h = h,s
        elif self.is_hovering:
            i = [min(i[0]+6,255),min(i[1]+6,255),min(i[2]+6,255)]
            s = [min(s[0]+6,255),min(s[1]+6,255),min(s[2]+6,255)]
        
        return (
            h+h+h+h+
            h+h+h+h+
            h+s+h+h+
            h+h+h+h+
            i+i+i+i+
            s+s+s+s+
            h+s+h+h+
            s+s+s+s+
            s+s+s+s
            )
    s_oldshadow.__noautodoc__ = True
    def s_material(self,bg,o,i,s,h):
        if self.pressed:
            i = [max(bg[0]-20,0),max(bg[1]-20,0),max(bg[2]-20,0)]
        elif self.is_hovering:
            i = [max(bg[0]-10,0),max(bg[1]-10,0),max(bg[2]-10,0)]
        
        return (
            o+s+o+o+
            s+s+o+o+
            s+o+o+o+
            o+s+s+o+
            i+i+i+i+
            s+o+o+s+
            o+o+s+o+
            o+o+s+s+
            o+o+o+s
            )
    s_material.__noautodoc__ = True

class ButtonBorderWidgetLayer(BaseBorderWidgetLayer):
    """
    Subclass of :py:class:`BaseBorderWidgetLayer()` implementing Button-Style borders.
    
    This class is based on the :py:class:`ButtonBackground` class.
    This means that most styles are also available here and should look identical.
    
    Note that this class uses only 20 vertices and is thus not compatible with styles
    created for use with :py:class:`BaseBorderWidgetLayer`\ .
    
    Also note that the ``border`` argument also only receives two values instead of eight.
    """
    n_vertices = 20
    def __init__(self,name,widget,
                z_index=None,
                base_border=[0,0],base_offset=[0,0],
                border=[4,4],
                style="flat",
                batch=None, change_on_press=None,
                ):
        super(ButtonBorderWidgetLayer,self).__init__(name,widget,z_index,base_border,base_offset,border[0:2],style,batch,change_on_press)
    def genVertices(self):
        sx,sy, ex,ey = self.getPos()
        bx,by = self.bborder
        
        # Vertex Naming
        # Y
        # | 1       2
        # |  5     6
        # |  7     8
        # | 3       4
        # +-------> X
        # Quad order
        #    1
        # 4  5  2
        #    3
        # Within each quad, vertices are ordered from the bottom-left counter-clockwise
        
        v1 = sx,        ey
        v2 = ex,        ey
        v3 = sx,        sy
        v4 = ex,        sy
        
        v5 = sx+bx,     ey-by
        v6 = ex-bx,     ey-by
        v7 = sx+bx,     sy+by
        v8 = ex-bx,     sy+by
        
        q1 = v5+v6+v2+v1
        q2 = v8+v4+v2+v6
        q3 = v3+v4+v8+v7
        q4 = v3+v7+v5+v1
        q5 = v7+v8+v6+v5
        
        return q1+q2+q3+q4+q5
    genVertices.__noautodoc__ = True
    
    def s_flat(self,bg,o,i,s,h):
        # Flat style makes no difference between normal,hover and pressed
        cb1 = i+i+i+i
        cb2 = i+i+i+i
        cb3 = i+i+i+i
        cb4 = i+i+i+i
        cc  = i+i+i+i
        
        return cb1+cb2+cb3+cb4+cc
    s_flat.__noautodoc__ = True
    def s_gradient(self,bg,o,i,s,h):
        if self.pressed:
            i = s
        elif self.is_hovering:
            i = [min(i[0]+6,255),min(i[1]+6,255),min(i[2]+6,255)]
        cb1 = i+i+o+o
        cb2 = i+o+o+i
        cb3 = o+o+i+i
        cb4 = o+i+i+o
        cc  = i+i+i+i
        
        return cb1+cb2+cb3+cb4+cc
    s_gradient.__noautodoc__ = True
    def s_oldshadow(self,bg,o,i,s,h):
        if self.pressed:
            i = s
            s,h = h,s
        elif self.is_hovering:
            i = [min(i[0]+6,255),min(i[1]+6,255),min(i[2]+6,255)]
            s = [min(s[0]+6,255),min(s[1]+6,255),min(s[2]+6,255)]
        cb1 = h+h+h+h
        cb2 = s+s+s+s
        cb3 = s+s+s+s
        cb4 = h+h+h+h
        cc  = i+i+i+i
        
        return cb1+cb2+cb3+cb4+cc
    s_oldshadow.__noautodoc__ = True
    def s_material(self,bg,o,i,s,h):
        if self.pressed:
            i = [max(bg[0]-20,0),max(bg[1]-20,0),max(bg[2]-20,0)]
        elif self.is_hovering:
            i = [max(bg[0]-10,0),max(bg[1]-10,0),max(bg[2]-10,0)]
        cb1 = s+s+o+o
        cb2 = s+o+o+s
        cb3 = o+o+s+s
        cb4 = o+s+s+o
        cc  = i+i+i+i
        
        return cb1+cb2+cb3+cb4+cc
    s_material.__noautodoc__ = True
    
    def stretchColors(self,c):
        return c
    stretchColors.__noautodoc__ = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  __init__.py
#  
#  Copyright 2017 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = [
    "WatchingList",
    "register_pyglet_handler",
    "ActionDispatcher",
    "SmartRegistry",
    ]

import sys
import weakref
import threading

try:
    import bidict
except ImportError:
    HAVE_BIDICT = False
else:
    HAVE_BIDICT = True

from .gui import *

class WatchingList(list):
    """
    Subclass of :py:func:`list` implementing a watched list.
    
    A WatchingList will call the given callback with a reference to itself whenever it is modified.
    Internally, the callback is stored as a weak reference, meaning that the creator should keep a reference around.
    
    This class is used in :py:class:`peng3d.gui.widgets.BasicWidget()` to allow for modifying single coordinates of the pos and size properties.
    """
    def __init__(self,l,callback=None):
        if callback is not None:
            self.callback = weakref.WeakMethod(callback)
        else:
            self.callback = None
        super(WatchingList,self).__init__(l)
    def __setitem__(self,*args):
        super(WatchingList,self).__setitem__(*args)
        if self.callback is not None:
            c = self.callback()(self)

def register_pyglet_handler(peng,func,event,raiseErrors=False):
    """
    Registers the given pyglet-style event handler for the given pyglet event.
    
    This function allows pyglet-style event handlers to receive events bridged
    through the peng3d event system. Internally, this function creates a lambda
    function that decodes the arguments and then calls the pyglet-style event handler.
    
    The ``raiseErrors`` flag is passed through to the peng3d event system and will
    cause any errors raised by this handler to be ignored.
    
    .. seealso::
       See :py:meth:`~peng3d.peng.Peng.addEventListener()` for more information.
    """
    peng.addEventListener("pyglet:%s"%event,(lambda data:func(*data["args"])),raiseErrors)

class ActionDispatcher(object):
    """
    Helper Class to be used to enable action support.
    
    Actions are simple callbacks that are specific to the instance they are registered with.
    
    To be able to use actions, a class must be a subclass of :py:class:`ActionDispatcher()`\ .
    
    Creation of required data structures is handled automatically when the first action is added.
    
    Internally, this object uses the ``actions`` attribute to store a map of action names to
    a list of callbacks.
    """
    def addAction(self,action,func,*args,**kwargs):
        """
        Adds a callback to the specified action.
        
        All other positional and keyword arguments will be stored and passed to the function upon activation.
        """
        if not hasattr(self,"actions"):
            self.actions = {}
        if action not in self.actions:
            self.actions[action] = []
        self.actions[action].append((func,args,kwargs))
    def doAction(self,action):
        """
        Helper method that calls all callbacks registered for the given action.
        """
        if not hasattr(self,"actions"):
            return
        for f,args,kwargs in self.actions.get(action,[]):
            f(*args,**kwargs)


class SmartRegistry(object):
    """
    Smart registry allowing easy management of mappings from int to str and vice versa.
    
    Note that bidict is required to be able to use this class.
    
    ``data`` may be a dictionary to initialize the registry with. Only dictionaries
    gotten from the :py:attr:`data` property should be used.
    
    ``reuse_ids`` specifies whether or not the automatic ID generator should re-use
    old, now unused IDs. See :py:meth:`genNewID()` for more information.
    
    ``start_id`` is the lowest ID that will be generated by the automatic ID generator.
    
    ``max_id`` is the highest ID that will be generated by the automatic ID generator.
    Should this limit by reached, an :py:exc:`AssertionError` will be raised.
    
    ``default_reg`` may be a dictionary mapping IDs to names. It will only be used if
    ``data`` did not already contain a registry.
    
    It is possible to access the registry via the dict-style ``reg[key]`` notation.
    This will return the name of whatever object was used as the key.
    
    Registering is also possible in a similar manner, like ``reg[name]=id``\ .
    ``id`` may be ``None`` to automatically generate one.
    
    This class also supports the ``in`` operator, note that both IDs and names are checked.
    """
    def __init__(self,data=None,reuse_ids=False,start_id=0,max_id=float("inf"),default_reg=None):
        # TODO: fix max_id being a float by default
        assert HAVE_BIDICT
        
        self._data = data if data is not None else {}
        
        self.reuse_ids = reuse_ids
        # if true, new ids will be assigned from lowest available id
        # if false, an internal counter is used
        
        self.start_id = start_id
        self.max_id = max_id
        
        self.id_lock = threading.Lock()
        self.registry_lock = threading.Lock()
        
        if "reg" not in self._data:
            default_reg = default_reg if default_reg is not None else {}
            if not isinstance(default_reg,dict):
                raise TypeError("Default Registry must be a dictionary")
            # no reg yet, create a new one
            # ID->NAME mapping
            self._data["reg"] = default_reg
        for k in self._data["reg"].keys():
            if not isinstance(k,int):
                raise TypeError("All keys must be integers")
        for v in self._data["reg"].values():
            if not isinstance(v,str):
                raise TypeError("All values must be strings")
        self._data["reg"] = bidict.bidict(self._data["reg"])
        if not self.reuse_ids and "next_id" not in self._data:
            if self._data.get("reg",{}) != {}:
                # already data there, find highest id +1
                self._data["next_id"]=max(max(self._data["reg"].keys())+1,start_id)
            else:
                # no data there, use start_id
                self._data["next_id"]=self.start_id
    
    def genNewID(self):
        """
        Generates a new ID.
        
        If ``reuse_ids`` was false, the new ID will be read from an internal counter
        which is also automatically increased. This means that the newly generated ID is already reserved.
        
        If ``reuse_ids`` was true, this method starts counting up from ``start_id`` until it finds an
        ID that is not currently known. Note that the ID is not reserved, this means that
        calling this method simultaneously from multiple threads may cause the same ID to be returned twice.
        
        Additionally, if the ID is greater or equal to ``max_id``\ , an :py:exc:`AssertionError` is raised.
        """
        if self.reuse_ids:
            i = self.start_id
            while True:
                if i not in self._data["reg"]:
                    assert i<=self.max_id
                    return i # no need to change any variables
                i+=1
        else:
            with self.id_lock:
                # new id creation in lock, to avoid issues with multiple threads
                i = self._data["next_id"]
                assert i<=self.max_id
                self._data["next_id"]+=1
            return i
    
    def register(self,name,force_id=None):
        """
        Registers a name to the registry.
        
        ``name`` is the name of the object and must be a string.
        
        ``force_id`` can be optionally set to override the automatic ID generation
        and force a specific ID.
        
        Note that using ``force_id`` is discouraged, since it may cause problems when ``reuse_ids`` is false.
        """
        with self.registry_lock:
            if force_id is None:
                new_id = self.genNewID()
            else:
                new_id = force_id
            self._data["reg"][new_id]=name
            return new_id
    
    def normalizeID(self,in_id):
        """
        Takes in an object and normalizes it to its ID/integer representation.
        
        Currently, only integers and strings may be passed in, else a :py:exc:`TypeError`
        will be thrown.
        """
        if isinstance(in_id,int):
            assert in_id in self._data["reg"]
            return in_id
        elif isinstance(in_id,str):
            assert in_id in self._data["reg"].inv
            return self._data["reg"].inv[in_id]
        else:
            raise TypeError("Only int and str can be converted to IDs")
    def normalizeName(self,in_name):
        """
        Takes in an object and normalizes it to its name/string.
        
        Currently, only integers and strings may be passed in, else a :py:exc:`TypeError`
        will be thrown.
        """
        if isinstance(in_name,str):
            assert in_name in self._data["reg"].inv
            return in_name
        elif isinstance(in_name,int):
            assert in_name in self._data["reg"]
            return self._data["reg"][in_name]
        else:
            raise TypeError("Only int and str can be converted to names")
    
    @property
    def data(self):
        """
        Read-only property to access the internal data.
        
        This is a dictionary containing all information necessary to re-create the registry
        via the ``data`` argument.
        
        The returned object is fully JSON/YAML/MessagePack serializable, as it only contains
        basic python data types.
        """
        d = self._data
        d["reg"]=dict(d["reg"])
        return d
    
    def __getitem__(self,key):
        # to access registry as reg[obj] -> name
        return self.normalizeName(key)
    
    def __setitem(self,key,value):
        # None may be used as value for auto generation
        # to access registry as reg[name]=id
        self.register(key,value)
    
    def __contains__(self,value):
        return value in self._data["reg"] or value in self._data["reg"].inv



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  setup.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

# Distribution command:
# sudo python setup.py install sdist bdist register upload

# Test command:
# tox -- --x-display=$DISPLAY -v -m "'not graphical'"

#import peng3d.version as ver

import imp

def load_module(name):
    names = name.split(".")
    path = None
    for name in names:
        f, path, info = imp.find_module(name, path)
        path = [path]
    return imp.load_module(name, f, path[0], info)    

ver = load_module("peng3d.version")

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# Fix for very old python versions from https://docs.python.org/2/distutils/setupscript.html#additional-meta-data
# patch distutils if it can't cope with the "classifiers" or
# "download_url" keywords
from sys import version
if version < '2.2.3':
    from distutils.dist import DistributionMetadata
    DistributionMetadata.classifiers = None
    DistributionMetadata.download_url = None

try:
    longdesc = open("README.txt","r").read()
except Exception:
    longdesc = "Python and pyglet based 3D Engine and toolkit"

setup(name='peng3d',
      version=ver.VERSION,
      description="Python and pyglet based 3D Engine and toolkit", # from the github repo
      long_description=longdesc,
      author="notna",
      author_email="notna@apparat.org",
      url="https://github.com/not-na/peng3d",
      packages=['peng3d',"peng3d.actor","peng3d.gui","peng3d.util"],
      install_requires=[
          "pyglet>=1.3.2",
          "bidict>=0.14.2",
      ],
      provides=["peng3d"],
      setup_requires=['pytest-runner'],
      tests_require=['pytest'],
      classifiers=[
        "Development Status :: 5 - Production/Stable",
        
        "Environment :: MacOS X",
        "Environment :: Win32 (MS Windows)",
        "Environment :: X11 Applications",
        
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        
        "License :: OSI Approved",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        
        "Topic :: Games/Entertainment",
        "Topic :: Multimedia :: Graphics :: 3D Rendering",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        
        ],
      )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  resource.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = ["ResourceManager"]

import os

try:
    import pyglet
    from pyglet.gl import *
except ImportError:
    pass # Probably headless

try:
    import fastjson as json
except ImportError:
    import json
try:
    from .libs import jsoncomment as _jsoncomment
    json = _jsoncomment.JsonComment(json)
except ImportError:
    # try with a locally installed version, may not be fully compatible but worth a try
    try:
        import jsoncomment as _jsoncomment
        json = _jsoncomment.JsonComment(json)
    except ImportError:
        # not found at all, most features will still work, but comments and extraneous commas in JSON files will not
        pass

from . import model

class ResourceManager(object):
    """
    Manager that allows for efficient and simple loading and management of different kinds of resources.
    
    Currently supports textures and models out of the box, but extension is possible.
    
    Textures can be queried by any part of the application, they are only loaded on the first request and then cached for every request following it.
    
    The same caching and lazy-loading principle applies to models loaded via this system.
    """
    missingtexturename = "peng3d:missingtexture"
    def __init__(self,peng,basepath):
        self.basepath = basepath
        self.peng = peng
        
        maxsize = GLint()
        glGetIntegerv(GL_MAX_TEXTURE_SIZE,maxsize)
        maxsize = min(maxsize.value,self.peng.cfg["rsrc.maxtexsize"]) # This is here to avoid massive memory overhead when only loading a few textures
        self.texsize = maxsize
        
        self.categories = {}
        self.categoriesTexCache = {}
        self.categoriesTexBin = {}
        
        self.missingTexture = None
        
        self.modelcache = {}
        self.modelobjcache = {}
        
        self.peng.sendEvent("peng3d:rsrc.init",{"peng":self.peng,"rsrcMgr":self})
        
    def resourceNameToPath(self,name,ext=""):
        """
        Converts the given resource name to a file path.
        
        A resource path is of the format ``<app>:<cat1>.<cat2>.<name>`` where cat1 and cat2 can be repeated as often as desired.
        
        ``ext`` is the file extension to use, e.g. ``.png`` or similar.
        
        As an example, the resource name ``peng3d:some.category.foo`` with the extension ``.png`` results in the path ``<basepath>/assets/peng3d/some/category/foo.png``\ .
        
        This resource naming scheme is used by most other methods of this class.
        
        Note that it is currently not possible to define multiple base paths to search through.
        """
        nsplit = name.split(":")[1].split(".")
        return os.path.join(self.basepath,"assets",name.split(":")[0],*nsplit)+ext
    def resourceExists(self,name,ext=""):
        """
        Returns whether or not the resource with the given name and extension exists.
        
        This must not mean that the resource is meaningful, it simply signals that the file exists.
        """
        return os.path.exists(self.resourceNameToPath(name,ext))
    def addCategory(self,name):
        """
        Adds a new texture category with the given name.
        
        If the category already exists, it will be overridden.
        """
        self.categories[name]={}
        self.categoriesTexCache[name]={}
        self.categoriesTexBin[name]=pyglet.image.atlas.TextureBin(self.texsize,self.texsize)
        self.peng.sendEvent("peng3d:rsrc.category.add",{"peng":self.peng,"category":name})
    def getTex(self,name,category):
        """
        Gets the texture associated with the given name and category.
        
        ``category`` must have been created using :py:meth:`addCategory()` before.
        
        If it was loaded previously, a cached version will be returned.
        If it was not loaded, it will be loaded and inserted into the cache.
        
        See :py:meth:`loadTex()` for more information.
        """
        if category not in self.categoriesTexCache:
            return self.getMissingTex(category)
        if name not in self.categoriesTexCache[category]:
            self.loadTex(name,category)
        return self.categoriesTexCache[category][name]
    def loadTex(self,name,category):
        """
        Loads the texture of the given name and category.
        
        All textures currently must be PNG files, although support for more formats may be added soon.
        
        If the texture cannot be found, a missing texture will instead be returned. See :py:meth:`getMissingTexture()` for more information.
        
        Currently, all texture mipmaps will be generated and the filters will be set to
        :py:const:`GL_NEAREST` for the magnification filter and :py:const:`GL_NEAREST_MIPMAP_LINEAR` for the minification filter.
        This results in a pixelated texture and not  a blurry one.
        """
        try:
            img = pyglet.image.load(self.resourceNameToPath(name,".png"))
        except FileNotFoundError:
            img = self.getMissingTexture()
        texreg = self.categoriesTexBin[category].add(img)
        #texreg = texreg.get_transform(True,True) # Mirrors the image due to how pyglets coordinate system works
        # Strange behavior, sometimes needed and sometimes not
        self.categories[category][name]=texreg
        target = texreg.target
        texid = texreg.id
        texcoords = texreg.tex_coords
        # Prevents texture bleeding with texture sizes that are powers of 2, else weird lines may appear at certain angles.
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_LINEAR)
        glGenerateMipmap(GL_TEXTURE_2D)
        
        out = target,texid,texcoords
        self.categoriesTexCache[category][name]=out
        self.peng.sendEvent("peng3d:rsrc.tex.load",{"peng":self.peng,"name":name,"category":category})
        return out
    def getMissingTexture(self):
        """
        Returns a texture to be used as a placeholder for missing textures.
        
        A default missing texture file is provided in the assets folder of the source distribution.
        It consists of a simple checkerboard pattern of purple and black, this image may be copied to any project using peng3d for similar behavior.
        
        If this texture cannot be found, a pattern is created in-memory, simply a solid square of purple.
        
        This texture will also be cached separately from other textures.
        """
        if self.missingTexture is None:
            if self.resourceExists(self.missingtexturename,".png"):
                self.missingTexture = pyglet.image.load(self.resourceNameToPath(self.missingtexturename,".png"))
                return self.missingTexture
            else: # Falls back to create pattern in-memory
                self.missingTexture = pyglet.image.create(1,1,pyglet.image.SolidColorImagePattern([255,0,255,255]))
                return self.missingTexture
        else:
            return self.missingTexture
    def getMissingTex(self,cat):
        if cat not in self.categories:
            self.addCategory(cat)
        return self.loadTex(self.missingtexturename,cat)
    def addFromTex(self,name,img,category):
        """
        Adds a new texture from the given image.
        
        ``img`` may be any object that supports Pyglet-style copying in form of the ``blit_to_texture()`` method.
        
        This can be used to add textures that come from non-file sources, e.g. Render-to-texture.
        """
        texreg = self.categoriesTexBin[category].add(img)
        #texreg = texreg.get_transform(True,True) # Mirrors the image due to how pyglets coordinate system works
        # Strange behaviour, sometimes needed and sometimes not
        
        self.categories[category][name]=texreg
        target = texreg.target
        texid = texreg.id
        texcoords = texreg.tex_coords
        
        # Prevents texture bleeding with texture sizes that are powers of 2, else weird lines may appear at certain angles.
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_LINEAR)
        glGenerateMipmap(GL_TEXTURE_2D)
        
        out = target,texid,texcoords
        self.categoriesTexCache[category][name]=out
        return out
    
    def getModel(self,name):
        """
        Gets the model object by the given name.
        
        If it was loaded previously, a cached version will be returned.
        If it was not loaded, it will be loaded and inserted into the cache.
        """
        if name in self.modelobjcache:
            return self.modelobjcache[name]
        return self.loadModel(name)
    def loadModel(self,name):
        """
        Loads the model of the given name.
        
        The model will also be inserted into the cache.
        """
        m = model.Model(self.peng,self,name)
        self.modelobjcache[name]=m
        self.peng.sendEvent("peng3d:rsrc.model.load",{"peng":self.peng,"name":name})
        return m
    
    def getModelData(self,name):
        """
        Gets the model data associated with the given name.
        
        If it was loaded, a cached copy will be returned.
        It it was not loaded, it will be loaded and cached.
        """
        if name in self.modelcache:
            return self.modelcache[name]
        return self.loadModelData(name)
    def loadModelData(self,name):
        """
        Loads the model data of the given name.
        
        The model file must always be a .json file.
        """
        path = self.resourceNameToPath(name,".json")
        try:
            data = json.load(open(path,"r"))
        except Exception:
            # Temporary
            print("Exception during model load: ")
            import traceback;traceback.print_exc()
            return {}# will probably cause other exceptions later on, TODO
        
        out = {}
        
        if data.get("version",1)==1:
            # Currently only one version, basic future-proofing
            # This version should get incremented with breaking changes to the structure
            
            # Materials
            out["materials"]={}
            for name,matdata in data.get("materials",{}).items():
                m = model.Material(self,name,matdata)
                out["materials"][name]=m
            out["default_material"]=out["materials"][data.get("default_material",list(out["materials"].keys())[0])]
            
            # Bones
            out["bones"]={"__root__":model.RootBone(self,"__root__",{"start_rot":[0,0],"length":0})}
            for name,bonedata in data.get("bones",{}).items():
                b = model.Bone(self,name,bonedata)
                out["bones"][name]=b
            for name,bone in out["bones"].items():
                if name == "__root__":
                    continue
                bone.setParent(out["bones"][bone.bonedata["parent"]])
            
            # Regions
            out["regions"]={}
            for name,regdata in data.get("regions",{}).items():
                r = model.Region(self,name,regdata)
                r.material = out["materials"][regdata.get("material",out["default_material"])]
                r.bone = out["bones"][regdata.get("bone","__root__")]
                out["bones"][regdata.get("bone","__root__")].addRegion(r)
                out["regions"][name]=r
            
            # Animations
            out["animations"]={}
            out["animations"]["static"]=model.Animation(self,"static",{"type":"static","bones":{}})
            for name,anidata in data.get("animations",{}).items():
                a = model.Animation(self,name,anidata)
                a.setBones(out["bones"])
                out["animations"][name]=a
            out["default_animation"]=out["animations"][data.get("default_animation",out["animations"]["static"])]
        else:
            raise ValueError("Unknown version %s of model '%s'"%(data.get("version",1),name))
        
        self.modelcache[name]=out
        return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  __init__.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = ["GUIMenu","SubMenu","GUILayer"]

#from pprint import pprint
#import gc
#import weakref
#import sys
import collections

try:
    import pyglet
    from pyglet.gl import *
except ImportError:
    pass # Headless mode

from ..menu import Menu
from ..layer import Layer,Layer2D
from .widgets import *
from .button import *
from .slider import *
from .text import *
from .container import *
from .layered import *
from .. import util

class GUIMenu(Menu):
    """
    :py:class:`peng3d.menu.Menu` subclass adding 2D GUI Support.
    
    Note that widgets are not managed directly by this class, but rather by each :py:class:`SubMenu`\ .
    """
    def __init__(self,name,window,peng):
        super(GUIMenu,self).__init__(name,window,peng)
        pyglet.clock.schedule_interval(lambda dt: None,1./30)
        self.submenus = {}
        self.activeSubMenu = None
    
    def addSubMenu(self,submenu):
        """
        Adds a :py:class:`SubMenu` to this Menu.
        
        Note that nothing will be displayed unless a submenu is activated.
        """
        self.submenus[submenu.name] = submenu
    def changeSubMenu(self,submenu):
        """
        Changes the submenu that is displayed.
        
        :raises ValueError: if the name was not previously registered
        """
        if submenu not in self.submenus:
            raise ValueError("Submenu %s does not exist!"%submenu)
        elif submenu == self.activeSubMenu:
            return # Ignore double submenu activation to prevent bugs in submenu initializer
        old = self.activeSubMenu
        self.activeSubMenu = submenu
        if old is not None:
            self.submenus[old].on_exit(submenu)
            self.submenus[old].doAction("exit")
        self.submenu.on_enter(old)
        self.submenu.doAction("enter")
    
    def draw(self):
        """
        Draws each menu layer and the active submenu.
        
        Note that the layers are drawn first and may be overridden by the submenu and widgets.
        """
        super(GUIMenu,self).draw()
        self.submenu.draw()
    
    @property
    def submenu(self):
        """
        Property containing the :py:class:`SubMenu` instance that is currently active.
        """
        return self.submenus[self.activeSubMenu]

class SubMenu(util.ActionDispatcher):
    """
    Sub Menu of the GUI system.
    
    Each instance must be registered with their menu to work properly, see :py:meth:`GUIMenu.addSubMenu()`\ .
    
    Actions supported by default:
    
    ``enter`` is triggered everytime the :py:meth:`on_enter()` method has been called.
    
    ``exit`` is triggered everytime the :py:meth:`on_exit()` method has been called.
    """
    def __init__(self,name,menu,window,peng):
        self.name = name
        self.menu = menu
        self.window = window
        self.peng = peng
        
        self.widgets = collections.OrderedDict()
        
        self.bg = [242,241,240,255]
        self.bg_vlist = pyglet.graphics.vertex_list(4,
            "v2f",
            "c4B",
            )
        self.peng.registerEventHandler("on_resize",self.on_resize)
        self.on_resize(*self.size)
        
        self.batch2d = pyglet.graphics.Batch()
    
    def draw(self):
        """
        Draws the submenu and its background.
        
        Note that this leaves the OpenGL state set to 2d drawing.
        """
        # Sets the OpenGL state for 2D-Drawing
        self.window.set2d()
        
        # Draws the background
        if isinstance(self.bg,Layer):
            self.bg._draw()
        elif hasattr(self.bg,"draw") and callable(self.bg.draw):
            self.bg.draw()
        elif isinstance(self.bg,list) or isinstance(self.bg,tuple):
            self.bg_vlist.draw(GL_QUADS)
        elif callable(self.bg):
            self.bg()
        elif isinstance(self.bg,Background):
            # The background will be drawn via the batch
            if not self.bg.initialized:
                self.bg.init_bg()
                self.bg.redraw_bg()
                self.bg.initialized=True
        elif self.bg=="blank":
            pass
        else:
            raise TypeError("Unknown background type")
        
        # In case the background modified relevant state
        self.window.set2d()
        
        # Check that all widgets that need redrawing have been redrawn
        for widget in self.widgets.values():
            if widget.do_redraw:
                widget.on_redraw()
                widget.do_redraw = False
        
        # Actually draw the content
        self.batch2d.draw()
        
        # Call custom draw methods where needed
        for widget in self.widgets.values():
            widget.draw()
    
    def addWidget(self,widget):
        """
        Adds a widget to this submenu.
        """
        self.widgets[widget.name]=widget
    def getWidget(self,name):
        """
        Returns the widget with the given name.
        """
        return self.widgets[name]
    
    def delWidget(self,widget):
        """
        Deletes the widget by the given name.
        
        Note that this feature is currently experimental as there seems to be a memory leak with this method.
        """
        # TODO: fix memory leak upon widget deletion
        #print("*"*50)
        #print("Start delWidget")
        if isinstance(widget,BasicWidget):
            widget = widget.name
        
        if widget not in self.widgets:
            return
        
        w = self.widgets[widget]
        
        #print("refs: %s"%sys.getrefcount(w))
        
        w.delete()
        del self.widgets[widget]
        del widget
        
        #w_wref = weakref.ref(w)
        
        #print("GC: GARBAGE")
        #print(gc.garbage)
        #print("Widget Info")
        #print(w_wref())
        #import objgraph
        #print("Objgraph")
        #objgraph.show_refs([w], filename='./mem_widget.png')
        #print("refs: %s"%sys.getrefcount(w))
        #w_r = gc.get_referrers(w)
        #print("GC: REFS")
        #for w_ref in w_r:
        #    print(repr(w_ref)[:512])
        #print("GC: END")
        #print("len: %s"%len(w_r))
        #del w_ref,w_r
        #print("after del %s"%w_wref())
        #print("refs: %s"%sys.getrefcount(w))
        del w
        #print("Final WRef")
        #print(w_wref())
    
    def setBackground(self,bg):
        """
        Sets the background of the submenu.
        
        The background may be a RGB or RGBA color to fill the background with.
        
        Alternatively, a :py:class:`peng3d.layer.Layer` instance or other object with a ``.draw()`` method may be supplied.
        It is also possible to supply any other method or function that will get called.
        
        Also, the strings ``flat``\ , ``gradient``\ , ``oldshadow`` and ``material`` may be given, resulting in a background that looks similar to buttons. 
        
        Lastly, the string ``"blank"`` may be passed to skip background drawing.
        """
        self.bg = bg
        if isinstance(bg,list) or isinstance(bg,tuple):
            if len(bg)==3 and isinstance(bg,list):
                bg.append(255)
            self.bg_vlist.colors = bg*4
        elif bg in ["flat","gradient","oldshadow","material"]:
            self.bg = ContainerButtonBackground(self,borderstyle=bg)
            self.on_resize(self.window.width,self.window.height)
    
    @property
    def pos(self):
        return [0,0] # As property to prevent bug with accidental manipulation
    
    @property
    def size(self):
        return self.window.width,self.window.height
    
    def on_resize(self,width,height):
        sx,sy = width,height
        self.bg_vlist.vertices = [0,0, sx,0, sx,sy, 0,sy]
        if isinstance(self.bg,Background):
            if not self.bg.initialized:
                self.bg.init_bg()
                self.bg.initialized=True
            self.bg.redraw_bg()
    
    def on_enter(self,old):
        pass
    def on_exit(self,new):
        pass

class GUILayer(GUIMenu,Layer2D):
    """
    Hybrid of :py:class:`GUIMenu` and :py:class:`peng3d.layer.Layer2D`\ .
    
    This class allows you to create Head-Up Displays and other overlays easily.
    """
    # TODO: add safety recursion breaker if this is added as a layer to itself
    def __init__(self,name,menu,window,peng):
        Layer2D.__init__(self,menu,window,peng)
        GUIMenu.__init__(self,menu.name,window,peng)
    def draw(self):
        """
        Draws the Menu.
        """
        self.predraw()
        GUIMenu.draw(self)
        self.postdraw()

# Hack to allow Container to use the drawing method of SubMenu for itself
from ..gui import container
container.SubMenu = SubMenu

# To allow menus to subclass SubMenu
from .menus import *
